# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Import the necessary packages
import subprocess
import numpy as np
import sys
import joblib
import pandas as pd
from pandas_plink import read_plink1_bin

# Define the munging class
import genoml.dependencies

class harmonizing:
    def __init__(self, test_geno_prefix, test_out_prefix, ref_model_prefix, training_SNPs):
        
        # Initializing the variables we will use 
        self.test_geno_prefix = test_geno_prefix
        self.test_out_prefix = test_out_prefix
        self.ref_model_prefix = ref_model_prefix
        self.training_SNPs = training_SNPs

        infile_h5 = ref_model_prefix + ".dataForML.h5"
        self.df = pd.read_hdf(infile_h5, key = "dataForML")

    def generate_new_PLINK(self):        
        # Show first few lines of the dataframe
        print("")
        print("Your data looks like this (showing the first few lines of the left-most and right-most columns)...")
        print("#"*70)
        print(self.df.describe())
        print("#"*70)
        print("")

        # Save out and drop the PHENO and sample ID columns 
        y_test = self.df.PHENO
        X_test = self.df.drop(columns=['PHENO'])
        IDs_test = X_test.ID
        X_test = X_test.drop(columns=['ID'])

        # Save variables to use globally within the class 
        self.y_test = y_test
        self.X_test = X_test
        self.IDs_test = IDs_test

        # Read in the column of SNPs from the SNP+Allele file read in 
        snps_alleles_df = pd.read_csv(self.training_SNPs, header=None)
        snps_only = snps_alleles_df.iloc[:, 0]
        snps_temp = self.test_out_prefix + '.SNPS_only_toKeep_temp.txt'
        snps_only.to_csv(snps_temp, header=None, index=False)

        print(f"A temporary file of SNPs from your reference dataset to keep in your testing dataset has been exported here: {snps_temp}")

        # Prepare the bashes to keep the SNPs of interest from the reference dataset
        plink_exec = genoml.dependencies.check_plink()

        # Creating outfile with SNPs
        # Force the allele designations based on the reference dataset
        plink_outfile = self.test_out_prefix + ".refSNPs_andAlleles"
        
        print("")
        print(f"Now we will create PLINK binaries where the reference SNPS and alleles will be based off of your file here: {self.training_SNPs}")
        print("")

        bash1 = f"{plink_exec} --bfile " + self.test_geno_prefix + " --extract " + snps_temp + " --reference-allele " + self.training_SNPs + " --make-bed --out " + plink_outfile
        # Remove the .log file 
        bash2 = "rm " + plink_outfile + ".log"
        # Remove the .SNPS_only_toKeep_temp.txt file 
        bash3 = "rm " + snps_temp

        cmds_a = [bash1, bash2, bash3]

        for cmd in cmds_a:
            subprocess.run(cmd, shell=True)

        self.plink_outfile = plink_outfile

        print("")
        print(f"A new set of PLINK binaries generated from your test dataset with the SNPs you decided to keep from the reference dataset have been made here: {plink_outfile}")
        print("")

    # def read_PLINK(self):
    # # Read in using pandas PLINK (similar to munging)

    #     bed_file = self.plink_outfile + ".bed"
    #     plink_files_py = read_plink1_bin(bed_file)
    #     plink_files = plink_files_py.drop(['fid','father','mother','gender', 'trait', 'chrom', 'cm', 'pos','a1'])

    #     plink_files = plink_files.set_index({'sample':'iid','variant':'snp'})
    #     plink_files.values = plink_files.values.astype('int')

    #     # swap pandas-plink genotype coding to match .raw format...more about that below:

    #     # for example, assuming C in minor allele, alleles are coded in plink .raw labels homozygous for minor allele as 2 and homozygous for major allele as 0:
    #     #A A   ->    0   
    #     #A C   ->    1   
    #     #C C   ->    2
    #     #0 0   ->   NA

    #     # where as, read_plink1_bin flips these, with homozygous minor allele = 0 and homozygous major allele = 2
    #     #A A   ->    2   
    #     #A C   ->    1   
    #     #C C   ->    0
    #     #0 0   ->   NA

    #     two_idx = (plink_files.values == 2)
    #     zero_idx = (plink_files.values == 0)

    #     plink_files.values[two_idx] = 0
    #     plink_files.values[zero_idx] = 2

    #     plink_pd = plink_files.to_pandas()
    #     plink_pd.reset_index(inplace=True)
    #     raw_df = plink_pd.rename(columns={'sample': 'ID'})

    #     print("")
    #     print("Your data looks like this (showing the first few lines of the left-most and right-most columns)...")
    #     print("#"*70)
    #     print(raw_df.describe())
    #     print("#"*70)
    #     print("")

    #     self.raw_df = raw_df

    #     return raw_df

    def prep_refCols_file(self):
        # Make a list of the column names from the reference dataset
        ref_columns_list = self.df.columns.values.tolist()

        # Write out the columns to a text file we will use in munge later 
        ref_cols_outfile = self.test_out_prefix + ".refColsHarmonize_toKeep.txt"

        with open(ref_cols_outfile, 'w') as filehandle:
            for col in ref_columns_list:
                filehandle.write('%s\n' % col)

        print("")
        print(f"A file with the columns in the reference file, to later use in the munging step and keep these same columns for the test dataset, has been generated here: {ref_cols_outfile}")
        print("")

        return ref_columns_list 





# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pandas as pd
from sklearn import ensemble
from sklearn import feature_selection

class featureselection:
    def __init__(self, run_prefix, df, data_type, n_est):
        self.run_prefix = run_prefix
        self.featureRanks = None
        self.n_est = n_est
        self.data_type = data_type

        self.y = df['PHENO']
        self.X = df.drop(columns=['PHENO'])
        X = self.X
        self.IDs = X.ID
        self.X = X.drop(columns=['ID'])

    def rank(self):
        print(f"""
            Beginning featureSelection using {self.n_est} estimators...""")

        if (self.data_type == "d"):
            print(f"""
            using extraTrees Classifier for your discrete dataset 
            """)
            clf = ensemble.ExtraTreesClassifier(n_estimators=self.n_est)
        
        if (self.data_type == "c"):
            print(f"""
            using extraTrees Regressor for your continuous dataset
            """)
            clf = ensemble.ExtraTreesRegressor(n_estimators=self.n_est)
        
        clf.fit(self.X, self.y)
        self.featureRanks = clf.feature_importances_
        
        # Code to drop the features below threshold and return the data set like it was (aka add PHENO and IDs back)
        model = feature_selection.SelectFromModel(clf, prefit=True) # find this import at top
        df_editing = model.transform(self.X)
        print("""
        Printing feature name that corresponds to the dataframe column name, then printing the relative importance as we go...
        """)
        
        list_featureScores = []

        for col,score in zip(self.X.columns,clf.feature_importances_):
            print(col,score)
            list_featureScores.append([col, score])

        df_featureScores = pd.DataFrame(list_featureScores, columns=["Feature_Name", "Score"])
        #df_featureScores = df_featureScores[df_featureScores['Score'] !=0]
        df_featureScores = df_featureScores.sort_values(by=['Score'], ascending=False)
        featureScores_outfile = self.run_prefix + ".approx_feature_importance.txt"
        df_featureScores.to_csv(featureScores_outfile, index=False, sep="\t")

        print(f"""
        You have reduced your dataset to {df_editing.shape[0]} samples at {df_editing.shape[1]} features, not including ID and PHENO.
        """)

        y_df = self.y
        ID_df = pd.DataFrame(self.IDs)
        features_selected = model.get_support()
        X_reduced = self.X.iloc[:,features_selected]
        df_selecta = pd.concat([ID_df.reset_index(drop=True), y_df.reset_index(drop=True), X_reduced.reset_index(drop=True)], axis = 1, ignore_index=False)
        
        self.df_selecta = df_selecta
        self.featureScores_outfile = featureScores_outfile

        return df_selecta
    
    def export_data(self):
        ## Export reduced data
        outfile_h5 = self.run_prefix + ".dataForML.h5"
        self.df_selecta.to_hdf(outfile_h5, key='dataForML')

        features_list = self.df_selecta.columns.values.tolist()
    
        features_listpath = self.run_prefix + ".list_features.txt"
        with open(features_listpath, 'w') as f:
            for feature in features_list:
                f.write("%s\n" % feature)

        print(f"""Exporting a new {outfile_h5} file that has a reduced feature set based on your importance approximations. 
        This is a good dataset for general ML applications for the chosen PHENO as it includes only features that are likely to impact the model.

        An updated list of {len(features_list)} features, including ID and PHENO, that is in your munged dataForML.h5 file can be found here {features_listpath}
        
        A file with all your features, ranked from largest contributors at the top to smallest contributors at the bottom, can be found at {self.featureScores_outfile}.
        """)
        


# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Import the necessary packages 
import joblib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import sklearn
import sys
import xgboost
import numpy as np
from time import time
import statsmodels.formula.api as sm
from sklearn.metrics import explained_variance_score, mean_squared_error, median_absolute_error, r2_score

class test:
    def __init__(self, df, loaded_model, run_prefix):
        self.df = df 
        self.run_prefix = run_prefix
        self.loaded_model = loaded_model

    def prep_df(self):
        
        print("")
        print("Your data looks like this (showing the first few lines of the left-most and right-most columns)...")
        print("#"*70)
        print(self.df.describe())
        print("#"*70)
        print("")
        
        # Save out and drop the PHENO and sample ID columns 
        y_test = self.df.PHENO
        X_test = self.df.drop(columns=['PHENO'])
        IDs_test = X_test.ID
        X_test = X_test.drop(columns=['ID'])

        # Save variables to use globally within the class 
        self.y_test = y_test
        self.X_test = X_test
        self.IDs_test = IDs_test

        return X_test
    
    def performance_metrics(self):
        
        log_cols=["Explained_variance_score", "Mean_squared_error", "Median_absolute_error", "R2_score"]
        log_table = pd.DataFrame(columns=log_cols)
                
        self.loaded_model.fit(self.X_test, self.y_test)

        print("")
        print("#"*70)

        test_predictions = self.loaded_model.predict(self.X_test)
        test_predictions = test_predictions
        evs = explained_variance_score(self.y_test, test_predictions)
        print("Explained variance score: {:.4}".format(evs))
            
        test_predictions = self.loaded_model.predict(self.X_test)
        test_predictions = test_predictions
        mse = mean_squared_error(self.y_test, test_predictions)
        print("Mean squared error: {:.4}".format(mse))
            
        test_predictions = self.loaded_model.predict(self.X_test)
        test_predictions = test_predictions
        mae = median_absolute_error(self.y_test, test_predictions)
        print("Median absolute error: {:.4}".format(mae))
            
        test_predictions = self.loaded_model.predict(self.X_test)
        test_predictions = test_predictions
        r2s = r2_score(self.y_test, test_predictions)
        print("R^2 score: {:.4}".format(r2s))
            
        log_entry = pd.DataFrame([[evs, mse, mae, r2s]], columns=log_cols)
        log_table = log_table.append(log_entry)

        print("#"*70)

        print("")

        log_outfile = self.run_prefix + '.testedModel_allSamples_performanceMetrics.csv'

        print("")
        print(f"This table below is also logged as {log_outfile} and is in your current working directory...")
        print("#"*70)
        print(log_table)
        print("#"*70)
        print("")

        log_table.to_csv(log_outfile, index=False)

        self.log_table = log_table 
        return log_table
    
    def export_pheno_predictions(self):
        
        test_predicted_values = self.loaded_model.predict(self.X_test)
        test_predicted_values_df = pd.DataFrame(test_predicted_values)
        y_test_df = pd.DataFrame(self.y_test)
        IDs_test_df = pd.DataFrame(self.IDs_test)

        test_out = pd.concat([IDs_test_df.reset_index(), y_test_df.reset_index(drop=True), test_predicted_values_df.reset_index(drop=True)], axis = 1, ignore_index=True)
        test_out.columns=["INDEX","ID","PHENO_REPORTED","PHENO_PREDICTED"]
        test_out = test_out.drop(columns=["INDEX"])

        test_outfile = self.run_prefix + '.testedModel_allSample_predictions.csv'
        test_out.to_csv(test_outfile, index=False)

        print("")
        print(f"Preview of the exported predictions exported as {test_outfile}, these are pretty straight forward.")
        print("They generally include the sample ID, the previously reported phenotype, and the predicted phenotype from that algorithm.")
        print("#"*70)
        print(test_out.head())
        print("#"*70)

        self.test_out = test_out 
        return test_out
    
    def regression_summary(self):

        genoML_colors = ["cyan","purple"]

        sns_plot = sns.regplot(data=self.test_out, y="PHENO_REPORTED", x="PHENO_PREDICTED", scatter_kws={"color": "cyan"}, line_kws={"color": "purple"})

        plot_out = self.run_prefix + '.testedModel_allSamples_regressionPlot.png'
        sns_plot.figure.savefig(plot_out, dpi=600)

        print("")
        print(f"We are also exporting a regression plot for you here {plot_out}, this is a graphical representation of the difference between the reported and predicted phenotypes in the withheld test data for the best performing algorithm.")

        print("")
        print("Here is a quick summary of the regression comparing PHENO_REPORTED ~ PHENO_PREDICTED in the withheld test data...")
        print("")

        reg_model = sm.ols(formula='PHENO_REPORTED ~ PHENO_PREDICTED', data=self.test_out)
        fitted = reg_model.fit()
        print(fitted.summary())

        fitted_out = self.run_prefix + 'testedModel_allSamples_regressionSummary.csv'
        
        with open(fitted_out, 'w') as fh:
            fh.write(fitted.summary().as_csv())

        print(f"We are exporting this summary here: {fitted_out}")

        print("")
        print("...always good to see the P value for the predictor.")



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import argparse
import functools
import sys

from genoml import utils
from genoml import dependencies
from genoml.cli import continuous_supervised_train, continuous_supervised_tune, \
    continuous_supervised_test, discrete_supervised_train, \
    discrete_supervised_tune, discrete_supervised_test, munging, harmonizing


def handle_main():
    entry_points = [
        {"name": "continuous", "handler": handle_continuous,
        "description": "for processing continuous datatypes (ex: age at onset)"},
        {"name": "discrete", "handler": handle_discrete, 
        "description": "for processing discrete datatypes (ex: case vs. control status)"},
        {"name": "harmonize", "handler": handle_harmonize,
        "description": "for harmonizing incoming test datasets to use the same SNPs and reference alleles prior to munging, training, and testing"},
    ]
    handle_dispatcher(entry_points, "genoml", 0)


def handle_continuous():
    entry_points = [
        {"name": "supervised", "handler": handle_continuous_supervised,
         "description": "choice of munge, train, tune, or test your continuous dataset must be specified"},
    ]
    handle_dispatcher(entry_points, "genoml continuous", 1)


def handle_discrete():
    entry_points = [
        {"name": "supervised", "handler": handle_discrete_supervised,
         "description": "choice of munge, train, tune, or test your discrete dataset must be specified"},
    ]
    handle_dispatcher(entry_points, "genoml discrete", 1)


def handle_continuous_supervised():
    entry_points = [
        {"name": "munge", "handler": handle_continuous_supervised_munge,
         "description": "for supervised munging and preprocessing of your continuous dataset prior to training"},
        {"name": "train", "handler": handle_continuous_supervised_train,
         "description": "for supervised training of your munged continuous dataset by competing different algorithms"},
        {"name": "tune", "handler": handle_continuous_supervised_tune,
         "description": "for supervised tuning of your munged and trained continuous dataset"},
        {"name": "test", "handler": handle_continuous_supervised_test,
         "description": "for supervised testing of your model generated on continuous data on unseen continuous data (after its been harmonized and munged)"},
    ]
    handle_dispatcher(entry_points, "genoml continuous supervised", 2)


def handle_discrete_supervised():
    entry_points = [
        {"name": "munge", "handler": handle_discrete_supervised_munge,
         "description": "for supervised munging and preprocessing of your discrete dataset prior to training"},
        {"name": "train", "handler": handle_discrete_supervised_train,
         "description": "for supervised training of your munged discrete dataset by competing different algorithms"},
        {"name": "tune", "handler": handle_discrete_supervised_tune,
         "description": "for supervised tuning of your munged and trained discrete dataset"},
        {"name": "test", "handler": handle_discrete_supervised_test,
         "description": "for supervised testing of your model generated on discrete data on unseen discrete data (after its been harmonized and munged)"},
    ]
    handle_dispatcher(entry_points, "genoml discrete supervised", 2)


def handle_harmonize():
    handle_endpoints("genoml harmonize",
                     ["test_geno_prefix", "test_prefix", "ref_model_prefix",
                      "training_snps_alleles"],
                     harmonizing.main, 1)


def handle_continuous_supervised_munge():
    handle_endpoints("genoml discrete supervised munge",
                     ["prefix", "impute", "geno", "skip_prune", "pheno", "addit",
                      "feature_selection", "gwas", "p", "vif", "iter",
                      "ref_cols_harmonize", "umap_reduce", "target_columns", "confounders"],
                      functools.partial(munging.main, data_type="c"), 3)


def handle_continuous_supervised_train():
    handle_endpoints("genoml continuous supervised train",
                     ["prefix", "export_predictions", "matching_columns"],
                     continuous_supervised_train.main, 3)


def handle_continuous_supervised_tune():
    handle_endpoints("genoml continuous supervised tune",
                     ["prefix", "max_tune", "n_cv"],
                     continuous_supervised_tune.main, 3)


def handle_continuous_supervised_test():
    handle_endpoints("genoml continuous supervised test",
                     ["prefix", "test_prefix", "ref_model_prefix"],
                     continuous_supervised_test.main, 3)


def handle_discrete_supervised_munge():
    handle_endpoints("genoml discrete supervised munge",
                     ["prefix", "impute", "geno", "skip_prune", "pheno", "addit",
                      "feature_selection", "gwas", "p", "vif", "iter",
                      "ref_cols_harmonize", "umap_reduce", "target_columns", "confounders"],
                      functools.partial(munging.main, data_type="d"), 3)


def handle_discrete_supervised_train():
    handle_endpoints("genoml discrete supervised train",
                     ["prefix", "metric_max", "prob_hist", "auc",
                      "matching_columns"],
                     discrete_supervised_train.main, 3)


def handle_discrete_supervised_tune():
    handle_endpoints("genoml discrete supervised tune",
                     ["prefix", "metric_tune", "max_tune", "n_cv"],
                     discrete_supervised_tune.main, 3)


def handle_discrete_supervised_test():
    handle_endpoints("genoml discrete supervised test",
                     ["prefix", "test_prefix", "ref_model_prefix"],
                     discrete_supervised_test.main, 3)


def handle_dispatcher(entry_points, command_name, level):
    usage_description = f'{command_name} <command> [<args>]\n'
    for command in entry_points:
        usage_description += "   {name:15s} {description}\n".format(**command)

    parser = argparse.ArgumentParser(prog=command_name,
                                     description=f'{command_name}',
                                     usage=usage_description
                                     )

    parser.add_argument('command', help='Subcommand to run')
    args = parser.parse_args(sys.argv[level + 1:level + 2])

    candidates = []
    for command in entry_points:
        if command["name"] == args.command:
            command["handler"]()
            return
        if command["name"].startswith(args.command):
            candidates.append(command)

    if len(candidates) == 1:
        candidates[0]["handler"]()
        return

    parser.print_usage()
    print(f'Unrecognized command: {args.command}')
    exit(1)


def add_default_flag(parser, flag_name):
    if flag_name == "prefix":
        parser.add_argument('--prefix', type=str, default="GenoML_data",
                            help="Prefix for your output build.")

    elif flag_name == "metric_max":
        parser.add_argument('--metric_max', type=str, default='AUC',
                            choices=['AUC', "Balanced_Accuracy", "Specificity",
                                     "Sensitivity"],
                            help='How do you want to determine which algorithm'
                                 ' performed the best? [default: AUC].')

    elif flag_name == "verbose":
        parser.add_argument('-v', '--verbose', action='store_true',
                            default=False, help="Verbose output.")

    elif flag_name == "matching_columns":
        parser.add_argument('--matching_columns', type=str, default=None,
                            help="This is the list of intersecting columns "
                                 "between reference and testing datasets with "
                                 "the suffix *_finalHarmonizedCols_toKeep.txt")

    elif flag_name == "prob_hist":
        parser.add_argument('--prob_hist', type=bool, default=False)

    elif flag_name == "auc":
        parser.add_argument('--auc', type=bool, default=False)

    elif flag_name == "export_predictions":
        parser.add_argument('--export_predictions', type=bool, default=False)

    elif flag_name == "metric_tune":
        parser.add_argument('--metric_tune', type=str, default='AUC',
                            choices=['AUC', "Balanced_Accuracy"],
                            help='Using what metric of the best algorithm do '
                                 'you want to tune on? [default: AUC].')

    elif flag_name == "max_tune":
        parser.add_argument('--max_tune', type=int, default=50,
                            help='Max number of tuning iterations: (integer '
                                 'likely greater than 10). This governs the '
                                 'length of tuning process, run speed and the '
                                 'maximum number of possible combinations of '
                                 'tuning parameters [default: 50].')

    elif flag_name == "n_cv":
        parser.add_argument('--n_cv', type=int, default=5,
                            help='Number of cross validations: (integer likely '
                                 'greater than 3). Here we set the number of '
                                 'cross-validation runs for the algorithms '
                                 '[default: 5].')

    elif flag_name == "test_prefix":
        parser.add_argument('--test_prefix', type=str, default='GenoML_data',
                            help='Prefix for the dataset you would like to '
                                 'test against your reference model. '
                                 'Remember, the model will not function well '
                                 'if it does not include the same features, '
                                 'and these features should be on the same '
                                 'numeric scale, you can leave off the '
                                 '\'.dataForML.h5\' suffix.')

    elif flag_name == "ref_model_prefix":
        parser.add_argument('--ref_model_prefix', type=str,
                            default='GenoML_model',
                            help='Prefix of your reference model file, '
                                 'you can leave off the \'.joblib\' suffix.')

    elif flag_name == "test_geno_prefix":
        parser.add_argument('--test_geno_prefix', type=str,
                            default='genotype_binaries',
                            help='Prefix of the genotypes for the test '
                                 'dataset in PLINK binary format.',
                            required=True)

    elif flag_name == "training_snps_alleles":
        parser.add_argument('--training_snps_alleles', type=str,
                            default=None,
                            help='File to the SNPs and alleles file generated '
                                 'in the training phase that we will use to '
                                 'compare.',
                            required=True)

    elif flag_name == "pheno":
        parser.add_argument("--pheno", type=str, default=None,
                            help="Phenotype: (string file path). Path to CSV "
                                 "phenotype file [default: None].",
                            required=True)

    elif flag_name == "geno":
        parser.add_argument("--geno", type=str, default=None,
                            help="Genotype: (string file path). Path to PLINK "
                                 "format genotype file, everything before the "
                                 "*.bed/bim/fam [default: None].")

    elif flag_name == "skip_prune":
        parser.add_argument("--skip_prune", type=str, default="no",
                            help="[default: no].",
                            choices=["no", "yes"], required=False)
                            
    elif flag_name == "addit":
        parser.add_argument("--addit", type=str, default=None,
                            help="Additional: (string file path). Path to CSV "
                                 "format feature file [default: None].")

    elif flag_name == "gwas":
        parser.add_argument("--gwas", type=str, default=None,
                            help="GWAS summary stats: (string file path). "
                                 "Path to CSV format external GWAS summary "
                                 "statistics containing at least the columns "
                                 "SNP and P in the header [default: None].")

    elif flag_name == "p":
        parser.add_argument("--p", type=float, default=0.001,
                            help="P threshold for GWAS: (some value between "
                                 "0-1). P value to filter your SNP data on ["
                                 "default: 0.001].")

    elif flag_name == "vif":
        parser.add_argument("--vif", type=int, default=0,
                            help="Variance Inflation Factor (VIF): (integer). "
                                 "This is the VIF threshold for pruning "
                                 "non-genotype features. We recommend a value "
                                 "of 5-10. The default of 0 means no VIF "
                                 "filtering will be done. [default: 0].")

    elif flag_name == "iter":
        parser.add_argument("--iter", type=int, default=0,
                            help="Iterator: (integer). How many iterations of "
                                 "VIF pruning of features do you want to run. "
                                 "To save time VIF is run in randomly "
                                 "assorted chunks of 1000 features per "
                                 "iteration. The default of 1 means only one "
                                 "pass through the data. [default: 1].")

    elif flag_name == "impute":
        parser.add_argument("--impute", type=str, default="median",
                            help="Imputation: (mean, median). Governs "
                                 "secondary imputation and data "
                                 "transformation [default: median].",
                            choices=["median", "mean"])

    elif flag_name == "feature_selection":
        parser.add_argument('--feature_selection', type=int, default=0,
                            help='Run a quick tree-based feature selection '
                                 'routine prior to anything else, here you '
                                 'input the integer number of estimators '
                                 'needed, we suggest >= 50. The default of 0 '
                                 'will skip this functionality. This will '
                                 'also output a reduced dataset for analyses '
                                 'in addition to feature ranks. [default: 0]')

    elif flag_name == "ref_cols_harmonize":
        parser.add_argument('--ref_cols_harmonize', type=str, default=None,
                            help='Are you now munging a test dataset '
                                 'following the harmonize step? Here you '
                                 'input the path to the to the '
                                 '*_refColsHarmonize_toKeep.txt file '
                                 'generated at that step.')
    
    elif flag_name == "umap_reduce":
        parser.add_argument('--umap_reduce', type=str, default="no",
                            help = 'Would you like to reduce your dimensions with UMAP? [default: ', 
                            choices=["no", "yes"], required=False)
    #    if args.umap_reduce and (args.target_columns is None or args.confounders is None):
    #        parser.error("Using the --umap_reduce option requires the --target_columns and --confounders options as well")

    elif flag_name == "target_columns":
        parser.add_argument('--target_columns', type=str, default=None,
                            help = 'For UMAP reduction. A .txt file, one column, with a list of features '
                            'to adjust (no header). These should correspond to features '
                            'in the munged dataset', required=False)

    elif flag_name == "confounders":
        parser.add_argument('--confounders', type=str, default=None,
                            help = 'For UMAP reduction. A .csv of confounders to adjust for with ID column and header.'
                            'Numeric, with no missing data and the ID column '
                            'is mandatory', required=False)

    else:
        raise Exception(f"Unknown flag: {flag_name}")


def handle_endpoints(command_name, flag_names, endpoint, level):
    parser = argparse.ArgumentParser(prog=command_name)
    for name in flag_names:
        add_default_flag(parser, name)

    add_default_flag(parser, "verbose")

    args, unknown = parser.parse_known_args(sys.argv[level + 1:])
    if unknown:
        parser.print_usage()
        print(f'Unrecognized arguments: {unknown[0]}')
        exit(1)

    utils.ContextScope._verbose = args.verbose
    args = [args.__dict__[name] for name in flag_names]
    args_string = ", ".join([f"{name}: {value if value else '[None]'}" for name, value in zip(flag_names, args)])
    with utils.ContextScope(f"Running {command_name}",
                            description="Args= " + args_string,
                            error=""):
        dependencies.check_dependencies()
        endpoint(*args)


if __name__ == "__main__":
    handle_main()



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import joblib
import pandas as pd
import seaborn as sns
import statsmodels.formula.api as sm
import xgboost
from sklearn import ensemble
from sklearn import linear_model
from sklearn import metrics
from sklearn import model_selection
from sklearn import neighbors
from sklearn import neural_network
from sklearn import svm

from genoml import utils


class train:
    def __init__(self, df, run_prefix):
        y = df.PHENO
        x = df.drop(columns=['PHENO'])

        x_train, x_test, y_train, y_test = model_selection.train_test_split(x, y, test_size=0.3,
                                                                            random_state=42)  # 70:30
        ids_train = x_train.ID
        ids_test = x_test.ID
        x_train = x_train.drop(columns=['ID'])
        x_test = x_test.drop(columns=['ID'])

        self._df = df
        self._run_prefix = run_prefix
        self._x_train = x_train
        self._x_test = x_test
        self._y_train = y_train
        self._y_test = y_test
        self._ids_train = ids_train
        self._ids_test = ids_test

        self.log_table = None
        self.best_algorithm = None
        self.algorithm = None
        self.rfe_df = None

        candidate_algorithms = [
            ensemble.AdaBoostRegressor(),
            ensemble.BaggingRegressor(),
            ensemble.GradientBoostingRegressor(),
            ensemble.RandomForestRegressor(n_estimators=10),
            linear_model.LinearRegression(),
            linear_model.SGDRegressor(),
            neighbors.KNeighborsRegressor(),
            neural_network.MLPRegressor(),
            svm.SVR(gamma='auto'),
            xgboost.XGBRegressor()
        ]

        self._algorithms = {algorithm.__class__.__name__: algorithm for algorithm in candidate_algorithms}
        self._best_algorithm_name = None
        self._best_algorithm = None
        self._best_algorithm_metrics = None

    def summary(self):
        """Report and data summary you want"""
        utils.DescriptionLoader.print("continuous/supervised/training/Train/summary", data=self._df.describe())

    @utils.DescriptionLoader.function_description("continuous/supervised/training/Train/compete")
    def compete(self):
        """Compete the algorithms"""

        competing_metrics = [metrics.explained_variance_score, metrics.mean_squared_error,
                             metrics.median_absolute_error, metrics.r2_score]
        column_names = ["algorithm", "runtime_s"] + [metric.__name__ for metric in competing_metrics]

        results = []
        for algorithm_name, algorithm in self._algorithms.items():
            with utils.DescriptionLoader.context("continuous/supervised/training/Train/compete/algorithm",
                                                 name=algorithm_name):
                algorithm.fit(self._x_train, self._y_train)
                with utils.Timer() as timer:
                    test_predictions = algorithm.predict(self._x_test)
                    metric_results = [metric_func(self._y_test, test_predictions) for metric_func in competing_metrics]
                row = [algorithm_name, timer.elapsed()] + metric_results

                results_str = self.metrics_to_str(dict(zip(column_names, row)))
                with utils.DescriptionLoader.context("continuous/supervised/training/Train/compete/algorithm/results",
                                                     name=algorithm_name, results=results_str):
                    results.append(row)

        self.log_table = pd.DataFrame(data=results, columns=column_names)

        best_id = self.log_table.explained_variance_score.idxmax()
        self._best_algorithm_name = self.log_table.iloc[best_id].algorithm
        self._best_algorithm = self._algorithms[self._best_algorithm_name]
        self._best_algorithm_metrics = self.log_table.iloc[best_id].to_dict()

        utils.DescriptionLoader.print("continuous/supervised/training/Train/compete/algorithm/best",
                                      algorithm=self._best_algorithm_name,
                                      metrics=self.metrics_to_str(self._best_algorithm_metrics))

    @staticmethod
    def metrics_to_str(metrics_dict):
        rows = []
        for key, value in metrics_dict.items():
            if key == "algorithm":
                rows.append("{}: {}".format(key, value))
            elif key == "runtime_s":
                rows.append("{}: {:0.3f} seconds\n".format(key, value))
            else:
                rows.append("{}: {:0.4f}".format(key, value))
        return str.join("\n", rows)

    def export_model(self):
        output_path = self._run_prefix + '.trainedModel.joblib'
        with utils.DescriptionLoader.context("continuous/supervised/training/Train/export_model",
                                             output_path=output_path):
            joblib.dump(self._best_algorithm, output_path)

    def export_predictions(self):
        output_columns = ["ID", "PHENO_REPORTED", "PHENO_PREDICTED"]

        train_predicted_values = self._best_algorithm.predict(self._x_train)
        results = pd.DataFrame(zip(self._ids_train, self._y_train, train_predicted_values), columns=output_columns)
        output_path = self._run_prefix + '.trainedModel_trainingSample_Predictions.csv'

        with utils.DescriptionLoader.context("continuous/supervised/training/Train/export_predictions/train_data",
                                             output_path=output_path, data=results.head()):
            results.to_csv(output_path, index=False)

        test_predicted_values = self._best_algorithm.predict(self._x_test)
        results = pd.DataFrame(zip(self._ids_test, self._y_test, test_predicted_values), columns=output_columns)
        output_path = self._run_prefix + '.trainedModel_withheldSample_Predictions.csv'

        with utils.DescriptionLoader.context("continuous/supervised/training/Train/export_predictions/test_data",
                                             output_path=output_path, data=results.head()):
            results.to_csv(output_path, index=False)

        output_path = self._run_prefix + '.trainedModel_withheldSample_regression.png'
        reg_model = sm.ols(formula='PHENO_REPORTED ~ PHENO_PREDICTED', data=results)
        fitted = reg_model.fit()
        with utils.DescriptionLoader.context("continuous/supervised/training/Train/export_predictions/plot",
                                             output_path=output_path, data=fitted.summary()):
            sns_plot = sns.regplot(data=results, y="PHENO_REPORTED", x="PHENO_PREDICTED", scatter_kws={"color": "cyan"},
                                   line_kws={"color": "purple"})

            sns_plot.figure.savefig(output_path, dpi=600)

    def save_algorithm_results(self, output_prefix):
        output_path = output_prefix + '.training_withheldSamples_performanceMetrics.csv'
        with utils.DescriptionLoader.context("continuous/supervised/training/Train/save_algorithm_results",
                                             output_path=output_path,
                                             data=self.log_table.describe()):
            self.log_table.to_csv(output_path, index=False)

    def save_best_algorithm(self, output_prefix):
        output_path = output_prefix + '.best_algorithm.txt'
        with utils.DescriptionLoader.context("continuous/supervised/training/Train/save_best_algorithm",
                                             output_path=output_path, best_algorithm=self._best_algorithm_name):
            with open(output_path, 'w') as fp:
                fp.write(self._best_algorithm_name)



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pandas as pd

from genoml.continuous import supervised


def main(run_prefix, max_iter, cv_count):
    # TUNING 
    # Create a dialogue with the user 
    print("Here is some basic info on the command you are about to run.")
    print("CLI argument info...")
    print(f"Working with the dataset and best model corresponding to prefix {run_prefix} the timestamp from the merge is the prefix in most cases.")
    print(f"Your maximum number of tuning iterations is {max_iter} and if you are concerned about runtime, make this number smaller.")
    print(f"You are running {cv_count} rounds of cross-validation, and again... if you are concerned about runtime, make this number smaller.")
    print("Give credit where credit is due, for this stage of analysis we use code from the great contributors to python packages: argparse, xgboost, sklearn, pandas, numpy, time, matplotlib and seaborn.")
    print("As a note, in all exported probabilities and other graphics, case status is treated as a 0 or 1, with 1 representing a positive case.")

    print("")
    
    infile_h5 = run_prefix + ".dataForML.h5"
    df = pd.read_hdf(infile_h5, key = "dataForML")

    y_tune = df.PHENO
    X_tune = df.drop(columns=['PHENO'])
    IDs_tune = X_tune.ID
    X_tune = X_tune.drop(columns=['ID'])


    best_algo_name_in = run_prefix + '.best_algorithm.txt'
    best_algo_df = pd.read_csv(best_algo_name_in, header=None, index_col=False)
    best_algo = str(best_algo_df.iloc[0,0])
    

    # Communicate to the user the best identified algorithm 
    print(f"From previous analyses in the training phase, we've determined that the best algorithm for this application is {best_algo}... so let's tune it up and see what gains we can make!")

    # Tuning 
    ## This calls on the functions made in the tune class (tuning.py) at the genoml.continuous.supervised
    model_tune = supervised.tune(df, run_prefix, max_iter, cv_count)
    model_tune.select_tuning_parameters() # Returns algo, hyperparameters, and scoring_metric
    model_tune.apply_tuning_parameters() # Randomized search with CV to tune
    model_tune.report_tune() #  Summary of the top 10 iterations of the hyperparameter tune
    model_tune.summarize_tune() # Summary of the cross-validation 
    model_tune.compare_performance() # Compares tuned performance to baseline to 
    model_tune.export_tuned_data() # Export the newly tuned predictions 
    model_tune.export_tune_regression() # Export the tuned and fitted regression model 

    print("")
    print("End of tuning stage with GenoML.")
    print("")



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse
import sys

import genoml.dependencies
from genoml import preprocessing


def main(prefix, impute, geno, skip_prune, pheno, addit, feature_selection, gwas, p, vif, iter, ref_cols_harmonize, umap_reduce, target_columns, confounders, data_type):
    genoml.dependencies.check_dependencies()

    run_prefix = prefix
    impute_type = impute
    geno_path = geno
    prune_choice = skip_prune
    pheno_path = pheno
    addit_path = addit
    n_est = feature_selection
    gwas_path = gwas
    p_gwas = p
    vif_thresh = vif
    vif_iter = iter
    refColsHarmonize = ref_cols_harmonize
    umap_reduce = umap_reduce
    target_columns = target_columns
    confounders = confounders

    # Print configurations
    print("")
    print("Here is some basic info on the command you are about to run.")
    print("Python version info...")
    print(sys.version)
    print("CLI argument info...")
    print(
        f"The output prefix for this run is {run_prefix} and will be appended to later runs of GenoML.")
    print(f"Working with genotype data? {geno_path}")
    print(f"Do you want GenoML to prune your SNPs for you? {prune_choice}")
    print(f"Working with additional predictors? {addit_path}")
    print(f"Where is your phenotype file? {pheno_path}")
    print(f"Any use for an external set of GWAS summary stats? {gwas_path}")
    print(
        f"If you plan on using external GWAS summary stats for SNP filtering, we'll only keep SNPs at what P value? {p_gwas}")
    print(f"How strong is your VIF filter? {vif_thresh}")
    print(f"How many iterations of VIF filtering are you doing? {vif_iter}")
    print(
        f"The imputation method you picked is using the column {impute_type} to fill in any remaining NAs.")
    print(f"Will you be adjusting additional features using UMAP dimensionality reduction? {umap_reduce}")
    print(
        "Give credit where credit is due, for this stage of analysis we use code from the great contributors to python packages: os, sys, argparse, numpy, pandas, joblib, math and time. We also use PLINK v1.9 from https://www.cog-genomics.org/plink/1.9/.")
    print("")

    # Run the munging script in genoml.preprocessing
    munger = preprocessing.munging(pheno_path=pheno_path, run_prefix=run_prefix, impute_type=impute_type, skip_prune=prune_choice,
                     p_gwas=p_gwas, addit_path=addit_path, gwas_path=gwas_path, geno_path=geno_path, refColsHarmonize=refColsHarmonize)

    # Process the PLINK inputs (for pruning)
    df = munger.plink_inputs()

    # Run the UMAP dimension reduction 
    if (umap_reduce == "yes"):
        print(f"Beginning to reduce using UMAP...")
        umap_df = preprocessing.adjuster(run_prefix, df, target_columns, confounders, umap_reduce)
        df = umap_df.umap_reducer()

    # Run the feature selection using extraTrees
    if n_est > 0:
        featureSelection_df = preprocessing.featureselection(run_prefix, df, data_type, n_est)
        df = featureSelection_df.rank()
        featureSelection_df.export_data()

    # Run the VIF calculation
    if vif_iter > 0:
        vif_calc = preprocessing.vif(vif_iter, vif_thresh, df, 100, run_prefix)
        vif_calc.vif_calculations()

    # Thank the user
    print("Thank you for munging with GenoML!")



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import sys

import numpy as np
import pandas as pd

from genoml import utils
from genoml.continuous import supervised


# TODO(mary): use or remove export_predictions
@utils.DescriptionLoader.function_description("cli/continuous_supervised_train")
def main(run_prefix, export_predictions, matching_columns_path):
    utils.DescriptionLoader.print("cli/continuous_supervised_train/info",
                                  python_version=sys.version, prefix=run_prefix)

    input_path = run_prefix + ".dataForML.h5"
    with utils.DescriptionLoader.context(
            "cli/continuous_supervised_train/input", path=input_path):
        df = pd.read_hdf(input_path, key="dataForML")

    if matching_columns_path:
        with utils.DescriptionLoader.context(
                "cli/continuous_supervised_train/matching_columns_path",
                matching_columns_path=matching_columns_path):
            with open(matching_columns_path, 'r') as matchingCols_file:
                matching_column_names_list = matchingCols_file.read().splitlines()

            df = df[np.intersect1d(df.columns, matching_column_names_list)]

    model = supervised.train(df, run_prefix)
    model.summary()
    model.compete()
    model.export_model()
    model.export_predictions()
    model.save_algorithm_results(run_prefix)
    model.save_best_algorithm(run_prefix)



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pandas as pd
import numpy as np
import statsmodels.api as sm 
import statsmodels.formula.api as smf
import statistics
import umap.umap_ as umap
from joblib import dump, load 
import matplotlib.pyplot as plt
from matplotlib import style
import seaborn as sns

class adjuster:
    def __init__(self, run_prefix, df, target_columns, confounders, umap_reduce):
        self.run_prefix = run_prefix
        self.umap_reduce = umap_reduce
        self.target_columns = target_columns
        self.confounders = confounders
        
        df = self.run_prefix + ".dataForML.h5"
        self.munged_data = df

    def umap_reducer(self):
        target_data_df = pd.read_hdf(self.munged_data, 'dataForML')
        target_column_df = pd.read_csv(self.target_columns, names=['TARGETS'])
        confounders_df = pd.read_csv(self.confounders)

        if (self.umap_reduce == "yes"):
            IDs = confounders_df['ID']
            IDs_df = pd.DataFrame(IDs) 
            to_umap = confounders_df.drop(columns=['ID']) 

            reducer = umap.UMAP(random_state=153) 
            embedding = reducer.fit_transform(to_umap)

            embedding1 = pd.DataFrame(embedding[:,0])
            embedding2 = pd.DataFrame(embedding[:,1])

            out_data = pd.concat([IDs_df.reset_index(), embedding1.reset_index(drop=True), embedding2.reset_index(drop=True)], axis=1, ignore_index=True)
            out_data.columns = ['INDEX', 'ID', 'UMAP_embedding1', "UMAP_embedding2"]
            out_data = out_data.drop(columns=['INDEX'])

            # Plot 
            print(f"Exporting UMAP plot...")
            fig, ax = plt.subplots(figsize=(12,10))
            plt.scatter(embedding[:,0], embedding[:,1], cmap="cool")
            plt.title("Data Reduction to 2 Dimensions by UMAP", fontsize=18)
            plot_out = self.run_prefix + '.umap_plot.png'
            plt.savefig(plot_out, dpi=600)

            print(f"The UMAP plot has been exported and can be found here: {plot_out}")
            
            out_file = self.runplot_out = self.run_prefix + '.umap_data_reduction.csv'
            out_data.to_csv(out_file, index=False)

            print(f"The reduced UMAP 2 dimensions per sample .csv file can be found here: {out_file}")

            exported_reducer = reducer.fit(to_umap)
            algo_out = self.runplot_out = self.run_prefix + '.umap_clustering.joblib'
            dump(exported_reducer, algo_out)

            print(f"The UMAP .joblib  file can be found here: {algo_out}")

            confounders_df = out_data
            target_list = list(target_column_df['TARGETS'])

            confounder_list = list(confounders_df.columns[1:])
            columns_to_keep_list = list(target_data_df.columns)

            adjustments_df = target_data_df.merge(confounders_df, how='inner', on='ID', suffixes=['', '_y'])

            formula_for_confounders = ' + '.join(confounder_list)
            normalize_switch = "yes"

            if normalize_switch == "yes":
                for target in target_list:
                    current_target = str(target)
                    print(f"Normalizing the following feature: {current_target}")
                    
                    current_formula = current_target + " ~ " + formula_for_confounders
                    print(current_formula)
                    
                    target_model = smf.ols(formula=current_formula, data=adjustments_df).fit()
                    
                    if (normalize_switch == 'yes'):
                        adjustments_df['temp'] = pd.to_numeric(target_model.resid)
                        print(type(adjustments_df['temp']))
                        mean_scalar = adjustments_df['temp'].mean()
                        sd_scalar = adjustments_df['temp'].std()
                        adjustments_df[current_target] = (adjustments_df['temp'] - mean_scalar)/sd_scalar
                        adjustments_df.drop(columns=['temp'], inplace=True)
                    else:
                        adjustments_df[current_target] = pd.to_numeric(target_model.resid)

            adjusted_df = adjustments_df[columns_to_keep_list]

            outfile_h5 = self.run_prefix + ".dataForML.h5"
            adjusted_df.to_hdf(outfile_h5, key='dataForML', mode='w')

            print(f"The adjusted dataframe following UMAP reduction and normalization : {outfile_h5}")

        return adjusted_df



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Import the necessary packages
import subprocess
import sys
import numpy as np
import pandas as pd
from pandas_plink import read_plink1_bin

# Define the munging class
import genoml.dependencies

class munging:
    def __init__(self, pheno_path, run_prefix="GenoML_data", impute_type="median", skip_prune="no", p_gwas=0.001, addit_path=None, gwas_path=None, geno_path=None, refColsHarmonize=None):
        self.pheno_path = pheno_path
        self.run_prefix = run_prefix
        
        self.impute_type = impute_type
        self.p_gwas = p_gwas
        
        self.addit_path = addit_path
        self.gwas_path = gwas_path
        self.geno_path = geno_path

        self.skip_prune = skip_prune
        
        self.refColsHarmonize = refColsHarmonize

        # Reading in the phenotype file 
        self.pheno_df = pd.read_csv(pheno_path, engine='c')
        
        # Raise an error and exit if the phenotype file is not properly formatted
        try:
            if set(['ID','PHENO']).issubset(self.pheno_df.columns) == False:
                raise ValueError("""
                Error: It doesn't look as though your phenotype file is properly formatted. 
                Did you check that the columns are 'ID' and 'PHENO' and that controls=0 and cases=1?""")
        except ValueError as ve:
            print(ve)
            sys.exit()

        # Typecase to read in the ID column as a string and the PHENO as an integer
        self.pheno_df['ID'] = self.pheno_df['ID'].astype(str)
        self.pheno_df['PHENO'] = self.pheno_df['PHENO'].astype(int)

        if (addit_path==None):
            print("No additional features as predictors? No problem, we'll stick to genotypes.")
            self.addit_df = None
        else:
            self.addit_df = pd.read_csv(addit_path, engine='c')

        if (gwas_path==None):
            print("So you don't want to filter on P values from external GWAS? No worries, we don't usually either (if the dataset is large enough).")
            self.gwas_df = None
        else:
            self.gwas_df = pd.read_csv(gwas_path, engine='c')
            
        if (geno_path==None):
            print("So no genotypes? Okay, we'll just use additional features provided for the predictions.")
        else:
            print("Exporting genotype data")

    def plink_inputs(self):
        # Initializing some variables
        plink_exec = genoml.dependencies.check_plink()
        impute_type = self.impute_type
        addit_df = self.addit_df
        pheno_df = self.pheno_df

        outfile_h5 = self.run_prefix + ".dataForML.h5"
        pheno_df.to_hdf(outfile_h5, key='pheno', mode='w')

        if (self.geno_path != None):
            if (self.skip_prune == "no"):
            # Set the bashes
                bash1a = f"{plink_exec} --bfile " + self.geno_path + " --indep-pairwise 1000 50 0.1"
                bash1b = f"{plink_exec} --bfile " + self.geno_path + " --extract " + self.run_prefix + \
                    ".p_threshold_variants.tab" + " --indep-pairwise 1000 50 0.1"
            # may want to consider outputting temp_genos to dir in run_prefix
                bash2 = f"{plink_exec} --bfile " + self.geno_path + \
                    " --extract plink.prune.in --make-bed --out temp_genos"
                bash3 = "cut -f 2,5 temp_genos.bim > " + \
                    self.run_prefix + ".variants_and_alleles.tab"
                bash4 = "rm plink.log"
                bash5 = "rm plink.prune.*"
            #    bash6 = "rm " + self.run_prefix + ".log"
            # Set the bash command groups
                cmds_a = [bash1a, bash2, bash3, bash4, bash5]
                cmds_b = [bash1b, bash2, bash3, bash4, bash5]
            

                if (self.gwas_path != None) & (self.geno_path != None):
                    p_thresh = self.p_gwas
                    gwas_df_reduced = self.gwas_df[['SNP', 'p']]
                    snps_to_keep = gwas_df_reduced.loc[(
                        gwas_df_reduced['p'] <= p_thresh)]
                    outfile = self.run_prefix + ".p_threshold_variants.tab"
                    snps_to_keep.to_csv(outfile, index=False, sep="\t")
                    print(
                        f"Your candidate variant list prior to pruning is right here: {outfile}.")

                if (self.gwas_path == None) & (self.geno_path != None):
                    print(
                        f"A list of pruned variants and the allele being counted in the dosages (usually the minor allele) can be found here: {self.run_prefix}.variants_and_alleles.tab")
                    for cmd in cmds_a:
                        subprocess.run(cmd, shell=True)

                if (self.gwas_path != None) & (self.geno_path != None):
                    print(
                        f"A list of pruned variants and the allele being counted in the dosages (usually the minor allele) can be found here: {self.run_prefix}.variants_and_alleles.tab")
                    for cmd in cmds_b:
                        subprocess.run(cmd, shell=True)
        
            if (self.skip_prune == "yes"):
                bash1a = f"{plink_exec} --bfile " + self.geno_path 
                bash1b = f"{plink_exec} --bfile " + self.geno_path + " --extract " + self.run_prefix + ".p_threshold_variants.tab"
            # may want to consider outputting temp_genos to dir in run_prefix
                bash2 = f"{plink_exec} --bfile " + self.geno_path + " --make-bed --out temp_genos"
                bash3 = "cut -f 2,5 temp_genos.bim > " + self.run_prefix + ".variants_and_alleles.tab"
                bash4 = "rm plink.log"

            # Set the bash command groups
                cmds_a = [bash1a, bash2, bash3, bash4]
                cmds_b = [bash1b, bash2, bash3, bash4]

                if (self.gwas_path != None) & (self.geno_path != None):
                    p_thresh = self.p_gwas
                    gwas_df_reduced = self.gwas_df[['SNP', 'p']]
                    snps_to_keep = gwas_df_reduced.loc[(
                        gwas_df_reduced['p'] <= p_thresh)]
                    outfile = self.run_prefix + ".p_threshold_variants.tab"
                    snps_to_keep.to_csv(outfile, index=False, sep="\t")
                    print(
                        f"Your candidate variant list is right here: {outfile}.")

                if (self.gwas_path == None) & (self.geno_path != None):
                    print(
                        f"A list of variants and the allele being counted in the dosages (usually the minor allele) can be found here: {self.run_prefix}.variants_and_alleles.tab")
                    for cmd in cmds_a:
                        subprocess.run(cmd, shell=True)

                if (self.gwas_path != None) & (self.geno_path != None):
                    print(
                        f"A list of variants and the allele being counted in the dosages (usually the minor allele) can be found here: {self.run_prefix}.variants_and_alleles.tab")
                    for cmd in cmds_b:
                        subprocess.run(cmd, shell=True)

        if (self.geno_path != None):
            
            g = read_plink1_bin('temp_genos.bed')
            g_pruned = g.drop(['fid','father','mother','gender', 'trait', 'chrom', 'cm', 'pos','a1'])

            g_pruned = g_pruned.set_index({'sample':'iid','variant':'snp'})
            g_pruned.values = g_pruned.values.astype('int')

        # swap pandas-plink genotype coding to match .raw format...more about that below:

        # for example, assuming C in minor allele, alleles are coded in plink .raw labels homozygous for minor allele as 2 and homozygous for major allele as 0:
        #A A   ->    0   
        #A C   ->    1   
        #C C   ->    2
        #0 0   ->   NA

        # where as, read_plink1_bin flips these, with homozygous minor allele = 0 and homozygous major allele = 2
        #A A   ->    2   
        #A C   ->    1   
        #C C   ->    0
        #0 0   ->   NA

            two_idx = (g_pruned.values == 2)
            zero_idx = (g_pruned.values == 0)

            g_pruned.values[two_idx] = 0
            g_pruned.values[zero_idx] = 2

            g_pd = g_pruned.to_pandas()
            g_pd.reset_index(inplace=True)
            raw_df = g_pd.rename(columns={'sample': 'ID'})
        #    del raw_df.index.name
        #    del raw_df.columns.name
            
        # now, remove temp_genos
            bash_rm_temp = "rm temp_genos.*"
            print(bash_rm_temp)
            subprocess.run(bash_rm_temp, shell=True)

    # Checking the impute flag and execute
        # Currently only supports mean and median
        impute_list = ["mean", "median"]
        
        if (self.geno_path != None):

            if impute_type not in impute_list:
                return "The 2 types of imputation currently supported are 'mean' and 'median'"
            elif impute_type.lower() == "mean":
                raw_df = raw_df.fillna(raw_df.mean())
            elif impute_type.lower() == "median":
                raw_df = raw_df.fillna(raw_df.median())
            print("")
            print(
                f"You have just imputed your genotype features, covering up NAs with the column {impute_type} so that analyses don't crash due to missing data.")
            print("Now your genotype features might look a little better (showing the first few lines of the left-most and right-most columns)...")
            print("#"*70)
            print(raw_df.describe())
            print("#"*70)
            print("")

    # Checking the imputation of non-genotype features

        if (self.addit_path != None):
            if impute_type not in impute_list:
                return "The 2 types of imputation currently supported are 'mean' and 'median'"
            elif impute_type.lower() == "mean":
                addit_df = addit_df.fillna(addit_df.mean())
            elif impute_type.lower() == "median":
                addit_df = addit_df.fillna(addit_df.median())
            print("")
            print(
                f"You have just imputed your non-genotype features, covering up NAs with the column {impute_type} so that analyses don't crash due to missing data.")
            print("Now your non-genotype features might look a little better (showing the first few lines of the left-most and right-most columns)...")
            print("#"*70)
            print(addit_df.describe())
            print("#"*70)
            print("")

            # Remove the ID column
            cols = list(addit_df.columns)
            cols.remove('ID')
            addit_df[cols]

            # Z-scale the features
            print(f"Now Z-scaling your non-genotype features...")

            # Remove any columns with a standard deviation of zero
            print(f"Removing any columns that have a standard deviation of 0 prior to Z-scaling...")
            
            if any(addit_df.std() == 0.0):
                print("")
                print(f"Looks like there's at least one column with a standard deviation of 0. Let's remove that for you...")
                print("") 
                addit_keep = addit_df.drop(addit_df.std()[addit_df.std() == 0.0].index.values, axis=1)
                addit_keep_list = list(addit_keep.columns.values)
                
                addit_df = addit_df[addit_keep_list]
                
                addit_keep_list.remove('ID')
                removed_list = np.setdiff1d(cols, addit_keep_list)
                for removed_column in range(len(removed_list)):
                    print(f"The column {removed_list[removed_column]} was removed")
                    
                cols = addit_keep_list
            
            print("")
            for col in cols:
                if (addit_df[col].min() == 0.0) and (addit_df[col].max() == 1.0):
                    print(col, "is likely a binary indicator or a proportion and will not be scaled, just + 1 all the values of this variable and rerun to flag this column to be scaled.")
                else:
                    addit_df[col] = (addit_df[col] - addit_df[col].mean())/addit_df[col].std(ddof=0)
            
            print("")
            print("You have just Z-scaled your non-genotype features, putting everything on a numeric scale similar to genotypes.")
            print("Now your non-genotype features might look a little closer to zero (showing the first few lines of the left-most and right-most columns)...")
            print("#"*70)
            print(addit_df.describe())
            print("#"*70)

    # Saving out the proper HDF5 file
        if (self.geno_path != None):
            merged = raw_df.to_hdf(outfile_h5, key='geno')

        if (self.addit_path != None):
            merged = addit_df.to_hdf(outfile_h5, key='addit')

        if (self.geno_path != None) & (self.addit_path != None):
            pheno = pd.read_hdf(outfile_h5, key="pheno")
            geno = pd.read_hdf(outfile_h5, key="geno")
            addit = pd.read_hdf(outfile_h5, key="addit")
            temp = pd.merge(pheno, addit, on='ID', how='inner')
            merged = pd.merge(temp, geno, on='ID', how='inner')

        if (self.geno_path != None) & (self.addit_path == None):
            pheno = pd.read_hdf(outfile_h5, key="pheno")
            geno = pd.read_hdf(outfile_h5, key="geno")
            merged = pd.merge(pheno, geno, on='ID', how='inner')

        if (self.geno_path == None) & (self.addit_path != None):
            pheno = pd.read_hdf(outfile_h5, key="pheno")
            addit = pd.read_hdf(outfile_h5, key="addit")
            merged = pd.merge(pheno, addit, on='ID', how='inner')

        # Checking the reference column names flag 
        # If this is a step that comes after harmonize, then a .txt file with columns to keep should have been produced
        # This is a list of column names from the reference dataset that the test dataset was harmonized against 
        # We want to compare apples to apples, so we will only keep the column names that match
        if (self.refColsHarmonize != None):
            print("") 
            print(f"Looks like you are munging after the harmonization step. Great! We will keep the columns generated from your reference dataset from that harmonize step that was exported to this file: {self.refColsHarmonize}")
            print("")
            with open(self.refColsHarmonize, 'r') as refCols_file:
                ref_column_names_list = refCols_file.read().splitlines()
            
            # Keep the reference columns from the test dataset if found in test data
            matching_cols = merged[np.intersect1d(merged.columns, ref_column_names_list)]
            
            # Make a list of final features that will be included in the model 
                # This will be used again when remunging the reference dataset
            matching_cols_list = matching_cols.columns.values.tolist()

            # Save out the final list 
            intersecting_cols_outfile = self.run_prefix + ".finalHarmonizedCols_toKeep.txt"

            with open(intersecting_cols_outfile, 'w') as filehandle:
                for col in matching_cols_list:
                    filehandle.write('%s\n' % col)

            print(f"A final list of harmonized columns between your reference and test dataset has been generated here: {intersecting_cols_outfile}")
            print(f"Use this to re-train your reference dataset in order to move on to testing.")

            # Replace the dataframe variable with the matching options
            merged = matching_cols

        self.merged = merged
        merged.to_hdf(outfile_h5, key='dataForML')

        features_list = merged.columns.values.tolist()
    
        features_listpath = self.run_prefix + ".list_features.txt"
        with open(features_listpath, 'w') as f:
            for feature in features_list:
                f.write("%s\n" % feature)

        print(f"An updated list of {len(features_list)} features, including ID and PHENO, that is in your munged dataForML.h5 file can be found here {features_listpath}")


        print("")
        print(f"Your .dataForML file that has been fully munged can be found here: {outfile_h5}")


        return merged



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import random
import numpy as np
import pandas as pd
import joblib
from statsmodels.stats import outliers_influence


# Define the VIF class to be used in munging
class vif:
    def __init__(self, iterations, vif_threshold, df, chunk_size, run_prefix):
        self.iterations = iterations
        self.threshold = vif_threshold
        self.df = df
        self.id = df['ID']
        self.pheno = df['PHENO']
        self.cleaned_df = None
        self.chunk_size = chunk_size
        self.run_prefix = run_prefix

    def check_df(self):
        """
        check_df takes in dataframe as an argument and strips it of missing values and non-numerical information.
        ### Arguments:
            df {pandas dataframe} -- A dataframe 
        ### Returns:
            cleaned_df {pandas dataframe} -- A cleaned dataframe with no NA values and only numerical values 
        """

        df = self.df

        print("Stripping erroneous space, dropping non-numeric columns...")
        df.columns = df.columns.str.strip()

        print("Drop any rows where at least one element is missing...")
        # Convert any infinite values to NaN prior to dropping NAs
        df.replace([np.inf, -np.inf], np.nan)
        df.dropna(how='any', inplace=True)

        print("Keeping only numerical columns...")
        int_cols = \
            df = df._get_numeric_data()

        print("Checking datatypes...")
        data_type = df.dtypes

        # Subset df to include only relevant numerical types
        int_cols = df.select_dtypes(include=["int", "int16", "int32", "int64", "float",
                                             "float16", "float32", "float64"]).shape[1]

        original_df = df
        self.original_df = original_df

        print("Sampling 100 rows at random to reduce memory overhead...")
        cleaned_df = df.sample(n=100).copy().reset_index()
        cleaned_df.drop(columns=["index"], inplace=True)

        print("Dropping columns that are not SNPs...")
        cleaned_df.drop(columns=['PHENO'], axis=1, inplace=True)
        print("Dropped!")

        print("Cleaned!")
        self.cleaned_df = cleaned_df
        return cleaned_df

    def randomize_chunks(self):
        chunk_size = self.chunk_size
        cleaned_df = self.cleaned_df

        """
        randomize_chunks takes in a cleaned dataframe's column names, randomizes them, 
        and spits out randomized, chunked dataframes with only SNPs for the VIF calculation later
        ### Arguments:
            cleaned_df {pandas dataframe} -- A cleaned dataframe 
            chunk_size {int} -- Desired size of dataframe chunked (default=100)
        ### Returns:
            list_chunked_dfs {list dfs} -- A cleaned, randomized list of dataframes with only SNPs as columns
        """

        print("Shuffling columns...")
        col_names_list = cleaned_df.columns.values.tolist()
        col_names_shuffle = random.sample(col_names_list, len(col_names_list))
        cleaned_df = cleaned_df[col_names_shuffle]
        print("Shuffled!")

        print("Generating chunked, randomized dataframes...")
        chunked_list = [col_names_shuffle[i * chunk_size:(i + 1) * chunk_size] for i in
                        range((len(col_names_shuffle) + chunk_size - 1) // chunk_size)]
        df_list = []
        for each_list in chunked_list:
            temp_df = cleaned_df[each_list].astype(float)
            df_list.append(temp_df.copy())

        no_chunks = len(df_list)
        print(f"The number of dataframes you have moving forward is {no_chunks}")
        print("Complete!")
        return df_list

    def calculate_vif(self, df_list):
        """
        calculate_vif takes in an list of randomized dataframes and removes any variables
        that is greater than the specified threshold (default=5.0). This is to combat 
        multicolinearity between the variables. The function then returns a fully VIF-filtered
        dataframe.
        ### Arguments:
            df_list {list dfs} -- A list of cleaned, randomized pandas dataframes 
            threshold {float} -- Cut-off for dropping following the VIF calculation (default=5.0)
        ### Returns:
            glued_df {pandas df} -- A complete VIF-filtered dataframe 
        """
        threshold = self.threshold

        dropped = True
        print(f"Dropping columns with a VIF threshold greater than {threshold}")

        for df in df_list:
            while dropped:
                # Loop until all variables in dataset have a VIF less than the threshold 
                variables = df.columns
                dropped = False
                vif = []

                # Changed to look at indexing 
                # Added simple joblib parallelization
                vif = joblib.Parallel(n_jobs=5)(
                    joblib.delayed(outliers_influence.variance_inflation_factor)(df[variables].values, df.columns.get_loc(var)) for var in
                    variables)

                max_vif = max(vif)

                if np.isinf(max_vif):
                    maxloc = vif.index(max_vif)
                    print(f'Dropping "{df.columns[maxloc]}" with VIF > {threshold}')
                    dropped = True

                if max_vif > threshold:
                    maxloc = vif.index(max_vif)
                    print(f'Dropping "{df.columns[maxloc]}" with VIF = {max_vif:.2f}')
                    df.drop([df.columns.tolist()[maxloc]], axis=1, inplace=True)
                    dropped = True

        print("\nVIF calculation on all chunks complete! \n")

        print("Gluing the dataframe back together...")
        glued_df = pd.concat(df_list, axis=1)
        print("Full VIF-filtered dataframe generated!")

        return glued_df

    def iterate(self, checked):
        """
        The iterate function specifies a number of times to iterate through the shuffling 
        and VIF filtering functions 
        ### Arguments:
            number {int} -- An integer specifying the number of iterations to perform (default=5)
        ### Returns:
            features_toKeep {list} -- A list of features to keep, extracted from the final iteration
        """

        iterations = self.iterations

        for iteration in range(iterations):
            print(f"""
                \n\n
                Iteration {iteration + 1}
                \n\n
                """)
            rando = self.randomize_chunks()
            checked = self.calculate_vif(rando)

        # When done, make list of features to keep 
        features_toKeep = checked.columns.values.tolist()

        print(f"""
        \n\n
            Iterations Complete!
        \n\n
        """)
        return features_toKeep, checked

    def vif_calculations(self):
        outfile_h5 = self.run_prefix + ".dataForML.h5"
        checked = self.check_df()
        features, complete_vif = self.iterate(checked)

        # Return the original dataframe with the features to keep 
        complete_vif_original_df = self.original_df[features]
        complete_vif_original_df['ID'] = self.id
        complete_vif_original_df['PHENO'] = self.pheno
        complete_vif_original_df.to_hdf(outfile_h5, key='dataForML')

        return complete_vif_original_df



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import sys
from genoml import preprocessing


def main(test_geno_prefix, test_prefix, ref_model_prefix,
         training_snps_alleles):
    # Print configurations
    print("Here is some basic info on the command you are about to run.")
    print("Python version info...")
    print(sys.version)
    print("CLI argument info...")
    print(f"You are importing test dataset {test_geno_prefix}.")
    print(
        f"Applying the model saved from your reference dataset in {ref_model_prefix}.")
    print(
        f"Reading in the SNP and allele information we will use to compare from {training_snps_alleles}.")
    print(
        f"The results of this test application of your model will be saved in files tagged {test_prefix}.")
    print(
        "As a note, in all exported probabilities and other graphics, case status is treated as a 0 or 1, with 1 representing a positive case.")

    # Run the harmonize script in genoml.preprocessing 
    harmonizer = preprocessing.harmonizing(test_geno_prefix=test_geno_prefix,
                                           test_out_prefix=test_prefix,
                                           ref_model_prefix=ref_model_prefix,
                                           training_SNPs=training_snps_alleles)

    # Generate new binaries from the test dataset using the reference dataset SNPs 
    harmonizer.generate_new_PLINK()

    # Read in PLINK binaries
    # harmonizer.read_PLINK()

    # Generate reference columns to keep for munging 
    harmonizer.prep_refCols_file()

    # Thank the user
    print("Thank you for harmonizing with GenoML!")



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Import the necessary packages 
import joblib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import sklearn
import sys
import xgboost
import numpy as np
from time import time
from sklearn.metrics import accuracy_score, balanced_accuracy_score, log_loss, roc_auc_score, confusion_matrix, roc_curve, auc

class test:
    def __init__(self, df, loaded_model, run_prefix):
        self.df = df 
        self.run_prefix = run_prefix
        self.loaded_model = loaded_model
    
    def prep_df(self):
        print("")
        print("Your data looks like this (showing the first few lines of the left-most and right-most columns)...")
        print("#"*70)
        print(self.df.describe())
        print("#"*70)
        print("")

        # Save out and drop the PHENO and sample ID columns 
        y_test = self.df.PHENO
        X_test = self.df.drop(columns=['PHENO'])
        IDs_test = X_test.ID
        X_test = X_test.drop(columns=['ID'])

        # Save variables to use globally within the class 
        self.y_test = y_test
        self.X_test = X_test
        self.IDs_test = IDs_test

        return X_test
    
    def export_ROC(self):
        
        # Define the output prefix  
        plot_out = self.run_prefix + '.testedModel_allSample_ROC.png'

        test_predictions = self.loaded_model.predict_proba(self.X_test)
        test_predictions = test_predictions[:, 1]

        fpr, tpr, thresholds = roc_curve(self.y_test, test_predictions)
        # Resolving issue #13 - ROC curve reporting to be consistent with performance metrics. 
        #roc_auc = auc(fpr, tpr)
        roc_auc = roc_auc_score(self.y_test, test_predictions)

        plt.figure()
        plt.plot(fpr, tpr, color='purple', label='All sample ROC curve (area = %0.2f)' % roc_auc + '.')
        plt.plot([0, 1], [0, 1], color='cyan', linestyle='--', label='Chance (area = %0.2f)' % 0.5)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver Operating Characteristic (ROC) - Test Dataset' )
        plt.legend(loc="lower right")
        plt.savefig(plot_out, dpi = 600)

        print("")
        print(f"We are also exporting a ROC curve for you here {plot_out} this is a graphical representation of AUC in all samples for the best performing algorithm.")

        self.roc_auc = roc_auc

        return roc_auc
            

    def export_tested_data(self):
        test_predicteds_probs = self.loaded_model.predict_proba(self.X_test)
        test_case_probs = test_predicteds_probs[:, 1]
        test_predicted_cases = self.loaded_model.predict(self.X_test)

        test_case_probs_df = pd.DataFrame(test_case_probs)
        test_predicted_cases_df = pd.DataFrame(test_predicted_cases)
        y_test_df = pd.DataFrame(self.y_test)
        IDs_test_df = pd.DataFrame(self.IDs_test)

        test_out = pd.concat([IDs_test_df.reset_index(), y_test_df.reset_index(drop=True), test_case_probs_df.reset_index(drop=True), test_predicted_cases_df.reset_index(drop=True)], axis = 1, ignore_index=True)
        test_out.columns=['INDEX','ID',"CASE_REPORTED","CASE_PROBABILITY","CASE_PREDICTED"]
        test_out = test_out.drop(columns=['INDEX'])

        test_outfile = self.run_prefix + '.testedModel_allSample_predictions.csv'
        test_out.to_csv(test_outfile, index=False)

        print("")
        print(f"Preview of the exported predictions for the testing samples exported as {test_outfile} in the similar format as in the initial training phase of GenoML.")
        print("#"*70)
        print(test_out.head())
        print("#"*70)

        self.test_out = test_out 

        return test_out
    
    def export_histograms(self):
        genoML_colors = ["cyan","purple"]

        g = sns.FacetGrid(self.test_out, hue="CASE_REPORTED", palette=genoML_colors, legend_out=True,)
        g = (g.map(sns.distplot, "CASE_PROBABILITY", hist=True, rug=False))
        g.add_legend()

        plot_out = self.run_prefix + '.testedModel_allSample_probabilities.png'
        g.savefig(plot_out, dpi=600)

        print("")
        print(f"We are also exporting probability density plots to the file {plot_out} this is a plot of the probability distributions of being a case, stratified by case and control status for all samples.")
        print("")

    def additional_sumstats(self):
        self.loaded_model.fit(self.X_test, self.y_test)

        print("")
        print("#"*70)
        print("Some additional summary stats logging from your application of your model to the test dataset.")
        print("")

        test_predictions = self.loaded_model.predict_proba(self.X_test)
        test_predictions = test_predictions[:, 1]
        
        # Resolving issue #13 - ROC curve reporting to be consistent with performance metrics. 
        # rocauc = roc_auc_score(self.y_test, test_predictions)
        rocauc = self.roc_auc
        print("AUC: {:.4%}".format(rocauc))

        test_predictions = self.loaded_model.predict(self.X_test)
        acc = accuracy_score(self.y_test, test_predictions)
        print("Accuracy: {:.4%}".format(acc))

        test_predictions = self.loaded_model.predict(self.X_test)
        balacc = balanced_accuracy_score(self.y_test, test_predictions)
        print("Balanced Accuracy: {:.4%}".format(balacc))
            
        CM = confusion_matrix(self.y_test, test_predictions)
        TN = CM[0][0]
        FN = CM[1][0]
        TP = CM[1][1]
        FP = CM[0][1]
        sensitivity = TP/(TP+FN)
        specificity = TN/(TN+FP)
        PPV = TP/(TP+FP)
        NPV = TN/(TN+FN)
        
        test_predictions = self.loaded_model.predict_proba(self.X_test)
        ll = log_loss(self.y_test, test_predictions)
        print("Log Loss: {:.4}".format(ll))

        log_cols=["AUC_Percent", "Accuracy_Percent", "Balanced_Accuracy_Percent", "Log_Loss", "Sensitivity", "Specificity", "PPV", "NPV"]
        log_table = pd.DataFrame(columns=log_cols)
        log_entry = pd.DataFrame([[rocauc*100, acc*100, balacc*100, ll, sensitivity, specificity, PPV, NPV]], columns=log_cols)
        log_table = log_table.append(log_entry)
        
        print("")
        print("#"*70)
        print("")

        log_outfile = self.run_prefix + '.testedModel_allSamples_performanceMetrics.csv'

        print(f"This table below is also logged as {log_outfile} and is in your current working directory...")
        print("#"*70)
        print(log_table)
        print("#"*70)

        print("")

        log_table.to_csv(log_outfile, index=False)

        self.log_table = log_table

        return log_table


# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import sys
import numpy as np
import pandas as pd

from genoml.discrete import supervised


def main(prefix, metric_max, prob_hist, auc, matchingCols):
    print("")
    print("Here is some basic info on the command you are about to run.")
    print("Python Version info...")
    print(sys.version) 

    # Print out chosen CLI arguments 
    print("CLI argument info...")
    print(f"Working with dataset {prefix} from previous data munging efforts.")
    print(f"You have chosen to compete the algorithms based on {metric_max}.")
    print("Give credit where credit is due, for this stage of analysis we use code from the great contributors to Python packages: argparse, xgboost, sklearn, pandas, numpy, time, matplotlib and seaborn.")
    print("As a note, in all exported probabilities and other graphics, case status is treated as a 0 or 1, with 1 representing a positive case.")
    print("")

    # Specify prefix and dataframe variables to be passed into class
    run_prefix = prefix
    infile_h5 = run_prefix + ".dataForML.h5"
    df = pd.read_hdf(infile_h5, key = "dataForML")

    if (matchingCols != None):
        print(f"Looks like you are retraining your reference file. We are using the harmonized columns you provided here: {matchingCols}")
        print(f"Note that you might have different/less features than before, given this was harmonized between training and test dataset, and might mean your model now performs worse...")

        with open(matchingCols, 'r') as matchingCols_file:
                matching_column_names_list = matchingCols_file.read().splitlines()
        
        # Keep only the columns found in the file 
        df = df[np.intersect1d(df.columns, matching_column_names_list)]

    model = supervised.train(df, run_prefix)
    model.summary()
    
    # Give user context prior to competing algorithms 
    # Explains to users how we are splitting their data 70:30 
    print("")
    print("Now let's compete these algorithms!")
    print("We'll update you as each algorithm runs, then summarize at the end.")
    print("Here we test each algorithm under default settings using the same training and test datasets derived from a 70% training and 30% testing split of your data.")
    print("For each algorithm, we will output the following metrics...")
    print("Algorithm name, hoping that's pretty self-explanatory. Plenty of resources on these common ML algorithms at https://scikit-learn.org and https://xgboost.readthedocs.io/.")
    print("AUC_percent, this is the area under the curve from receiver operating characteristic analyses. This is the most common metric of classifier performance in biomedical literature, we express this as a percent. We calculate AUC based on the predicted probability of being a case.")
    print("Accuracy_percent, this is the simple accuracy of the classifier, how many predictions were correct from best classification cutoff (python default).")
    print("Balanced_Accuracy_Percent, consider this as the accuracy resampled to a 1:1 mix of cases and controls. Imbalanced datasets can give funny results for simple accuracy.")
    print("Log_Loss, this is essentially the inverse of the likelihood function for a correct prediction, you want to minimize this.")
    print("Sensitivity, proportion of cases correctly identified.")
    print("Specificity, proportion of controls correctly identified.")
    print("PPV, this is the positive predictive value, the probability that subjects with a positive result actually have the disease.")
    print("NPV, this is the negative predictive value, the probability that subjects with a negative result don't have the disease.")
    print("We also log the runtimes per algorithm.")
    print("")
    print("Algorithm summaries incoming...")
    print("")

    # Compete the algorithms 
    model.compete()
    
    # Output the results of the log
    model.results(metric_max)

    # Export the results 
    model.export_model()

    # Export the AUC     
    model.AUC(save=True)

    # Export the probability histograms
    model.export_prob_hist()

    # Save out the proper algorithm
    model.save_results(prefix, algorithmResults = True, bestAlgorithm = True)

    print("Thank you for training with GenoML!")


# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import os
import time
import traceback

__author__ = 'Sayed Hadi Hashemi'

import textwrap


class ColoredBox:
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39

    def __init__(self, color=None):
        if color is None:
            color = self.GREEN
        self.__color = color

    def __enter__(self):
        print('\033[{}m'.format(self.__color), end="")

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("\x1b[0m", end="")

    @classmethod
    def wrap(cls, text, color):
        return '\033[{}m'.format(color) + text + "\x1b[0m"


class ContextScope:
    indent = 0
    _verbose = False

    def __init__(self, title, description, error, start=True, end=False,
                 **kwargs):
        self._title = title.format(**kwargs)
        self._description = description.format(**kwargs)
        self._error = error.format(**kwargs)
        self._start = start
        self._end = end

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            if self._end:
                print(
                    "{}{}: {}".format(
                        self.get_prefix(ColoredBox.GREEN),
                        ColoredBox.wrap(self._title, ColoredBox.GREEN),
                        ColoredBox.wrap('[Done]', ColoredBox.GREEN)))
            self.remove_indent()
        else:
            print("{}{}: {}".format(
                self.get_prefix(ColoredBox.RED), self._title,
                ColoredBox.wrap('[Failed]', ColoredBox.RED)))
            print("{}".format(self.indent_text(self._error)))
            self.remove_indent()
            traceback.print_exception(exc_type, exc_val, exc_tb)
            exit(1)

    def __enter__(self):
        self.add_indent()
        if self._start:
            print()
            print("{}{}".format(self.get_prefix(ColoredBox.BLUE),
                                ColoredBox.wrap(self._title, ColoredBox.BLUE)))
        if self._verbose and self._description:
            print("{}".format(self._description))

    @classmethod
    def add_indent(cls):
        cls.indent += 1

    @classmethod
    def remove_indent(cls):
        cls.indent -= 1

    @classmethod
    def get_prefix(cls, color=None):
        indent_size = 4
        # text = "=" * (cls.indent * 4) + "> "
        text = "---> " * cls.indent
        if color:
            text = ColoredBox.wrap(text, color)
        return text

    @classmethod
    def indent_text(cls, text):
        WIDTH = 70
        indent = max(0, len(cls.get_prefix()) - 2)
        width = WIDTH - indent
        ret = textwrap.fill(text, width)
        ret = textwrap.indent(ret, " " * indent)
        return ret

    @classmethod
    def set_verbose(cls, verbose):
        cls._verbose = verbose


def function_description(**dkwargs):
    def wrap(func):
        def func_wrapper(*args, **kwargs):
            with ContextScope(**dkwargs):
                return func(*args, **kwargs)

        return func_wrapper

    return wrap


class DescriptionLoader:
    _descriptions = None

    @classmethod
    def _load(cls):
        description_file = os.path.join(os.path.dirname(__file__),
                                        "misc", "descriptions.json")
        with open(description_file) as fp:
            cls._descriptions = json.load(fp)

    @classmethod
    def function_description(cls, key, **kwargs):
        dkwargs = cls.get(key)
        return function_description(**dkwargs, **kwargs)

    @classmethod
    def get(cls, key):
        if cls._descriptions is None:
            cls._load()
        return cls._descriptions[key]

    @classmethod
    def context(cls, key, **kwargs):
        dkwargs = cls.get(key)
        return ContextScope(**dkwargs, **kwargs)

    @classmethod
    def print(cls, key, **kwargs):
        dkwargs = cls.get(key)
        with ContextScope(**dkwargs, **kwargs):
            pass


class Timer:
    def __init__(self):
        self.start = None
        self.end = None

    def start_timer(self):
        self.start = time.time()

    def __enter__(self):
        self.start_timer()
        return self

    def __exit__(self, *args):
        self.stop_timer()

    def stop_timer(self):
        self.end = time.time()

    def elapsed(self):
        return self.end - self.start



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import setuptools

with open('requirements.txt') as file:
    requires = [line.strip() for line in file if not line.startswith('#')]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="genoml2",
    version="1.0.0-beta.5",
    maintainer="The GenoML Development Team",
    maintainer_email="genoml@googlegroups.com",
    description="GenoML is an automated machine learning tool that optimizes"
                " basic machine learning pipelines for genomic data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://genoml.github.io/",
    download_url="https://github.com/GenoML/genoml2/archive/v1.0.0-beta.5.tar.gz",
    entry_points={
        'console_scripts':
            ['genoml=genoml.__main__:handle_main'],
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'joblib', 
        'matplotlib',
        'numpy',
        'tables',
        'pandas',
        'pandas_plink',
        'requests',
        'scikit-learn',
        'scipy',
        'seaborn',
        'statsmodels',
        'xgboost',
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={'genoml': ['misc/*']},
)



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import logging
import os
import pathlib
import platform
import requests
import stat
import subprocess
import zipfile

from genoml import utils


def __get_executable_folder():
    key = "GENOML_DEP_DIR"
    if key in os.environ:
        return os.path.abspath(os.environ.get(key))
    else:
        return os.path.join(str(pathlib.Path.home()), ".genoml", "misc",
                            "executables")


__executable_folder = __get_executable_folder()


def __check_exec(exec_path, *args, absolute_path=False):
    if not absolute_path:
        binary_path = os.path.join(__executable_folder, exec_path)
    else:
        binary_path = exec_path
    if not os.path.exists(binary_path):
        return False

    _ = subprocess.run([binary_path, *args], stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL)
    return True


def __install_exec(url, exec_path):
    r = requests.get(url, verify=False, stream=True)
    r.raw.decode_content = True
    buffer = io.BytesIO()
    buffer.write(r.content)
    with zipfile.ZipFile(buffer, "r") as fp:
        fp.extractall(__executable_folder)

    binary_path = os.path.join(__executable_folder, exec_path)
    os.chmod(binary_path, stat.S_IEXEC)


def __check_package(name):
    platform_system = platform.system()

    if name not in __DEPENDENCIES:
        raise EnvironmentError("Unknown package: {}".format(name))

    if platform_system not in __DEPENDENCIES[name]:
        raise EnvironmentError(
            "Unknown supported OK: {}".format(platform_system))

    entry = __DEPENDENCIES[name][platform_system]

    binary_name = entry["binary"]
    args = entry["version_args"]
    url = entry["url"]

    if __check_exec(binary_name, *args):
        logging.debug("{} is found".format(name))
        return os.path.join(__executable_folder, binary_name)

    logging.warning("Installing {}".format(name))
    __install_exec(url, binary_name)
    if not __check_exec(binary_name, *args):
        logging.warning("Failed to run {} after installation".format(name))
        raise EnvironmentError("Can not install {}".format(name))
    else:
        return os.path.join(__executable_folder, binary_name)


@utils.DescriptionLoader.function_description("check_dependencies")
def check_dependencies():
    global __DEPENDENCIES
    ret = {}
    for package, data in __DEPENDENCIES.items():
        if "checker" in data:
            with utils.DescriptionLoader.context(
                    "check_dependencies_{}".format(package)):
                ret[package] = data["checker"]()

    return ret


def check_plink():
    return __check_package("Plink")


__DEPENDENCIES = {
    "Plink": {
        "checker": check_plink,
        "Darwin": {
            "binary": "plink",
            "version_args": ["--version"],
            "url": "http://s3.amazonaws.com/plink1-assets/plink_mac_20200219.zip"
        },
        "Linux": {
            "binary": "plink",
            "version_args": ["--version"],
            "url": "http://s3.amazonaws.com/plink1-assets/plink_linux_x86_64_20200219.zip"
        }
    },
}



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import joblib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats
import sklearn
from sklearn import discriminant_analysis
from sklearn import ensemble
from sklearn import linear_model
from sklearn import metrics
from sklearn import model_selection
from sklearn import neighbors
from sklearn import neural_network
from sklearn import svm
from time import time
import xgboost

class tune():
    def __init__(self, df, run_prefix, max_iter, cv_count):
        self.run_prefix = run_prefix
        self.max_iter = max_iter
        self.cv_count = cv_count
       
        self.y_tune = df.PHENO
        self.X_tune = df.drop(columns=['PHENO'])
        self.IDs_tune = self.X_tune.ID
        self.X_tune = self.X_tune.drop(columns=['ID'])

        best_algo_name_in = run_prefix + '.best_algorithm.txt'
        best_algo_df = pd.read_csv(best_algo_name_in, header=None, index_col=False)
        self.best_algo = str(best_algo_df.iloc[0,0])

        self.algorithms = [
        linear_model.LogisticRegression(),
        ensemble.RandomForestClassifier(),
        ensemble.AdaBoostClassifier(),
        ensemble.GradientBoostingClassifier(),
        linear_model.SGDClassifier(loss='modified_huber'),
        svm.SVC(probability=True),
        neural_network.MLPClassifier(),
        neighbors.KNeighborsClassifier(),
        discriminant_analysis.LinearDiscriminantAnalysis(),
        discriminant_analysis.QuadraticDiscriminantAnalysis(),
        ensemble.BaggingClassifier(),
        xgboost.XGBClassifier()
        ]
        self.log_table = None
        self.best_algo_name_in = None
        self.best_algo_df = None
        self.hyperparameters = None
        self.scoring_metric = None
        self.cv_tuned = None 
        self.cv_baseline = None 
        self.algo = None
        self.searchCVResults = None
        self.rand_search = None
        self.algo_tuned = None
        self.tune_out = None

    def select_tuning_parameters(self, metric_tune):
        best_algo = self.best_algo
        self.metric_tune = metric_tune 
        
        if best_algo == 'LogisticRegression':
            algo = getattr(sklearn.linear_model, best_algo)()

        elif  best_algo == 'SGDClassifier':
            algo = getattr(sklearn.linear_model, best_algo)(loss='modified_huber')

        elif (best_algo == 'RandomForestClassifier') or (best_algo == 'AdaBoostClassifier') or (best_algo == 'GradientBoostingClassifier') or  (best_algo == 'BaggingClassifier'):
            algo = getattr(sklearn.ensemble, best_algo)()

        elif best_algo == 'SVC':
            algo = getattr(sklearn.svm, best_algo)(probability=True, gamma='auto')

        elif best_algo == 'ComplementNB':
            algo = getattr(sklearn.naive_bayes, best_algo)()

        elif best_algo == 'MLPClassifier':
            algo = getattr(sklearn.neural_network, best_algo)()

        elif best_algo == 'XGBClassifier':
            algo = getattr(xgboost, best_algo)()

        elif best_algo == 'KNeighborsClassifier':
            algo = getattr(sklearn.neighbors, best_algo)()

        elif (best_algo == 'LinearDiscriminantAnalysis') or (best_algo == 'QuadraticDiscriminantAnalysis'):
            algo = getattr(sklearn.discriminant_analysis, best_algo)()

        self.algo = algo

        if(metric_tune == "AUC"):
            if best_algo == 'LogisticRegression':
                hyperparameters = {"penalty": ["l1", "l2"], "C": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            # Addressing issue #14 - SGDClassifier gets eta < 0 bug at tune
            elif best_algo == 'SGDClassifier':
                hyperparameters = { 'alpha': [1e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3], 'loss': ['log'], 'penalty': ['l2'], 'n_jobs': [-1] } 
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif (best_algo == 'RandomForestClassifier') or (best_algo == 'AdaBoostClassifier') or (best_algo == 'GradientBoostingClassifier') or  (best_algo == 'BaggingClassifier'):
                hyperparameters = {"n_estimators": stats.randint(1, 1000)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif best_algo == 'SVC':
                hyperparameters = {"kernel": ["linear", "poly", "rbf", "sigmoid"], "C": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)
                
            elif best_algo == 'ComplementNB':
                hyperparameters = {"alpha": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif best_algo == 'MLPClassifier':
                hyperparameters = {"alpha": stats.uniform(0,1), "learning_rate": ['constant', 'invscaling', 'adaptive']}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif best_algo == 'XGBClassifier':
                hyperparameters = {"max_depth": stats.randint(1, 100), "learning_rate": stats.uniform(0,1), "n_estimators": stats.randint(1, 100), "gamma": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif best_algo == 'KNeighborsClassifier':
                hyperparameters = {"leaf_size": stats.randint(1, 100), "n_neighbors": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)

            elif (best_algo == 'LinearDiscriminantAnalysis') or (best_algo == 'QuadraticDiscriminantAnalysis'):
                hyperparameters = {"tol": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.roc_auc_score, needs_proba=True)
            
        if(metric_tune == "Balanced_Accuracy"):
            if best_algo == 'LogisticRegression':
                hyperparameters = {"penalty": ["l1", "l2"], "C": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            # Addressing issue #14 - SGDClassifier gets eta < 0 bug at tune
            elif best_algo == 'SGDClassifier':
                hyperparameters = { 'alpha': [1e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3], 'loss': ['log'], 'penalty': ['l2'], 'n_jobs': [-1] } 
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif (best_algo == 'RandomForestClassifier') or (best_algo == 'AdaBoostClassifier') or (best_algo == 'GradientBoostingClassifier') or  (best_algo == 'BaggingClassifier'):
                hyperparameters = {"n_estimators": stats.randint(1, 1000)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif best_algo == 'SVC':
                hyperparameters = {"kernel": ["linear", "poly", "rbf", "sigmoid"], "C": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)
                
            elif best_algo == 'ComplementNB':
                hyperparameters = {"alpha": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif best_algo == 'MLPClassifier':
                hyperparameters = {"alpha": stats.uniform(0,1), "learning_rate": ['constant', 'invscaling', 'adaptive']}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif best_algo == 'XGBClassifier':
                hyperparameters = {"max_depth": stats.randint(1, 100), "learning_rate": stats.uniform(0,1), "n_estimators": stats.randint(1, 100), "gamma": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif best_algo == 'KNeighborsClassifier':
                hyperparameters = {"leaf_size": stats.randint(1, 100), "n_neighbors": stats.randint(1, 10)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)

            elif (best_algo == 'LinearDiscriminantAnalysis') or (best_algo == 'QuadraticDiscriminantAnalysis'):
                hyperparameters = {"tol": stats.uniform(0,1)}
                scoring_metric = metrics.make_scorer(metrics.balanced_accuracy_score, needs_proba=False)


        self.hyperparameters = hyperparameters
        self.scoring_metric = scoring_metric
            
        return algo, hyperparameters, scoring_metric

    def apply_tuning_parameters(self):
        # Randomized search with CV to tune
        print("Here is a summary of the top 10 iterations of the hyperparameter tuning...")
        rand_search = model_selection.RandomizedSearchCV(estimator=self.algo, param_distributions=self.hyperparameters, scoring=self.scoring_metric,n_iter=self.max_iter, cv=self.cv_count, n_jobs=-1, random_state=153, verbose=0)
        start = time()
        rand_search.fit(self.X_tune, self.y_tune)
        print("RandomizedSearchCV took %.2f seconds for %d candidates"
            " parameter iterations." % ((time() - start), self.max_iter))

        self.searchCVResults = rand_search.cv_results_
        self.rand_search = rand_search
        self.algo_tuned = rand_search.best_estimator_
        return rand_search.cv_results_

    def report_tune(self):
        # Summary of the top 10 iterations of the hyperparameter tune
        n_top=10
        results = self.searchCVResults
        
        top10_log_cols = ["Model_Rank", "Mean_Validation_Score", "Mean_Standard_Deviation", "Parameters"]
        top10_log_table = pd.DataFrame(columns=top10_log_cols)
        
        for i in range(1, n_top + 1):
            candidates = np.flatnonzero(results['rank_test_score'] == i)
            for candidate in candidates:
                print("Model with Rank: {0}".format(i))
                print("Mean Validation Score: {0:.3f} (std: {1:.3f})".format(
                results['mean_test_score'][candidate],
                results['std_test_score'][candidate]))
                print("Parameters: {0}".format(results['params'][candidate]))
                print("")
                top10_log_entry = pd.DataFrame([[i, results['mean_test_score'][candidate], results['std_test_score'][candidate], results['params'][candidate]]], columns=top10_log_cols)
                top10_log_table = top10_log_table.append(top10_log_entry)

        log_outfile = self.run_prefix + '.tunedModel_top10Iterations_Summary.csv'
        top10_log_table.to_csv(log_outfile, index=False)

        print(f"We are exporting a summary table of the top 10 iterations of the hyperparameter tuning step and its parameters here {log_outfile}.")

        return top10_log_table

    def summarize_tune(self):
    
        print("Here is the cross-validation summary of your best tuned model hyperparameters...")
        cv_tuned = model_selection.cross_val_score(estimator=self.rand_search.best_estimator_, X=self.X_tune, y=self.y_tune, scoring=self.scoring_metric, cv=self.cv_count, n_jobs=-1, verbose=0)
        print("Scores per cross-validation of the metric to be maximized, this scoring metric is AUC or Balanced_Accuracy for discrete phenotypes and explained variance for continuous phenotypes:")
        print(cv_tuned)
        print("Mean cross-validation score:")
        print(cv_tuned.mean())
        print("Standard deviation of the cross-validation score:")
        print(cv_tuned.std())

        print("")
        
        print("Here is the cross-validation summary of your baseline/default hyperparamters for the same algorithm on the same data...")
        cv_baseline = model_selection.cross_val_score(estimator=self.algo, X=self.X_tune, y=self.y_tune, scoring=self.scoring_metric, cv=self.cv_count, n_jobs=-1, verbose=0)
        print("Scores per cross-validation of the metric to be maximized, this scoring metric is AUC or Balanced_Accuracy for discrete phenotypes and explained variance for continuous phenotypes:")
        print(cv_baseline)
        print("Mean cross-validation score:")
        print(cv_baseline.mean())
        print("Standard deviation of the cross-validation score:")
        print(cv_baseline.std())

        print("")
        print("Just a note, if you have a relatively small variance among the cross-validation iterations, there is a higher chance of your model being more generalizable to similar datasets.")

        self.cv_baseline = cv_baseline
        self.cv_tuned = cv_tuned

        # Output a log table summarizing CV mean scores and standard deviations 
        summary_CV_log_cols = ["Mean_CV_Score_Baseline", "Standard_Dev_CV_Score_Baseline", "Min_CV_Score_Baseline", "Max_CV_Score_Baseline", "Mean_CV_Score_BestTuned", "Standard_Dev_CV_Score_BestTuned", "Min_CV_Score_BestTuned", "Max_CV_Score_BestTuned"]
        summary_CV_log_table = pd.DataFrame(columns=summary_CV_log_cols)
        summary_CV_log_entry = pd.DataFrame([[cv_baseline.mean(), cv_baseline.std(), cv_baseline.min(), cv_baseline.max(), cv_tuned.mean(), cv_tuned.std(), cv_tuned.min(), cv_tuned.max()]], columns=summary_CV_log_cols)
        summary_CV_log_table = summary_CV_log_table.append(summary_CV_log_entry)
        log_outfile = self.run_prefix + '.tunedModel_CV_Summary.csv'
        summary_CV_log_table.to_csv(log_outfile, index=False)

        print(f"We are exporting a summary table of the cross-validation mean score and standard deviation of the baseline vs. best tuned model here {log_outfile}.")

        return summary_CV_log_table

    def compare_performance(self):
        cv_tuned = self.cv_tuned
        cv_baseline = self.cv_baseline
        
        print("")
        if cv_baseline.mean() > cv_tuned.mean():
            print("Based on comparisons of the default parameters to your hyperparameter tuned model, the baseline model actually performed better.")
            print("Looks like the tune wasn't worth it, we suggest either extending the tune time or just using the baseline model for maximum performance.")
            print("")
            print("Let's shut everything down, thanks for trying to tune your model with GenoML.")

            return self.algo

        if cv_baseline.mean() < cv_tuned.mean():
            print("Based on comparisons of the default parameters to your hyperparameter tuned model, the tuned model actually performed better.")
            print("Looks like the tune was worth it, we suggest using this model for maximum performance, lets summarize and export this now.")
            print("In most cases, if opting to use the tuned model, a separate test dataset is a good idea. GenoML has a module to fit models to external data.")
            
            algo_tuned = self.rand_search.best_estimator_
            
            ### Save it using joblib
         
            algo_tuned_out = self.run_prefix + '.tunedModel.joblib'
            joblib.dump(algo_tuned, algo_tuned_out)

            return algo_tuned
    
    # def ROC(self):
    #     ### Export the ROC curve

    #     plot_out = self.run_prefix + '.tunedModel_allSample_ROC.png'

    #     test_predictions = self.algo_tuned.predict_proba(self.X_tune)
    #     test_predictions = test_predictions[:, 1]

    #     fpr, tpr, thresholds = metrics.roc_curve(self.y_tune, test_predictions)
    #     roc_auc = metrics.auc(fpr, tpr)

    #     plt.figure()
    #     plt.plot(fpr, tpr, color='purple', label='All sample ROC curve (area = %0.2f)' % roc_auc + '\nMean cross-validation ROC curve (area = %0.2f)' % self.cv_tuned.mean())
    #     plt.plot([0, 1], [0, 1], color='cyan', linestyle='--', label='Chance (area = %0.2f)' % 0.5)
    #     plt.xlim([0.0, 1.0])
    #     plt.ylim([0.0, 1.05])
    #     plt.xlabel('False positive rate')
    #     plt.ylabel('True positive rate')
    #     plt.title('Receiver operating characteristic (ROC) - ' + self.best_algo + '- tuned' )
    #     plt.legend(loc="lower right")
    #     plt.savefig(plot_out, dpi = 600)

    #     print()
    #     print(f"We are also exporting a ROC curve for you here {plot_out} this is a graphical representation of AUC in all samples for the best performing algorithm.")
    
    def export_tuned_data(self):
        tune_predicteds_probs = self.algo_tuned.predict_proba(self.X_tune)
        tune_case_probs = tune_predicteds_probs[:, 1]
        tune_predicted_cases = self.algo_tuned.predict(self.X_tune)

        tune_case_probs_df = pd.DataFrame(tune_case_probs)
        tune_predicted_cases_df = pd.DataFrame(tune_predicted_cases)
        y_tune_df = pd.DataFrame(self.y_tune)
        IDs_tune_df = pd.DataFrame(self.IDs_tune)

        tune_out = pd.concat([IDs_tune_df.reset_index(), y_tune_df.reset_index(drop=True), tune_case_probs_df.reset_index(drop=True), tune_predicted_cases_df.reset_index(drop=True)], axis = 1, ignore_index=True)
        tune_out.columns=['INDEX','ID',"CASE_REPORTED","CASE_PROBABILITY","CASE_PREDICTED"]
        tune_out = tune_out.drop(columns=['INDEX'])

        self.tune_out = tune_out

        tune_outfile = self.run_prefix + '.tunedModel_allSample_Predictions.csv'
        tune_out.to_csv(tune_outfile, index=False)

        print("")
        print(f"Preview of the exported predictions for the tuning samples which is naturally overfit and exported as {tune_outfile} in the similar format as in the initial training phase of GenoML.")
        print("#"*70)
        print(tune_out.head())
        print("#"*70)
    
    def export_tune_hist_prob(self):
        genoML_colors = ["cyan","purple"]

        sns_plot = sns.FacetGrid(self.tune_out, hue="CASE_REPORTED", palette=genoML_colors, legend_out=True)
        sns_plot = (sns_plot.map(sns.distplot, "CASE_PROBABILITY", hist=True, rug=False))
        sns_plot.add_legend()

        plot_out = self.run_prefix + '.tunedModel_allSample_probabilities.png'
        sns_plot.savefig(plot_out, dpi=600)

        print("")
        print(f"We are also exporting probability density plots to the file {plot_out} this is a plot of the probability distributions of being a case, stratified by case and control status for all samples.")



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import joblib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import sklearn
from sklearn import discriminant_analysis, ensemble, linear_model, metrics, model_selection, neighbors, neural_network, svm
import time
import xgboost


# Define the train class
class train:
    def __init__(self, df, run_prefix):
        #code that will prepare the data
        y = df.PHENO
        X = df.drop(columns=['PHENO'])
        
        # Split the data 
        X_train, X_test, y_train, y_test = model_selection.train_test_split(X, y, test_size=0.3, random_state=42) # 70:30
        IDs_train = X_train.ID
        IDs_test = X_test.ID
        X_train = X_train.drop(columns=['ID'])
        X_test = X_test.drop(columns=['ID'])

        # Saving the prepped data the other classes will need
        self.df = df
        self.run_prefix = run_prefix
        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test
        self.IDs_train = IDs_train
        self.IDs_test = IDs_test

        # Where the results will be stored 
        self.log_table = None
        self.best_algo = None
        self.algo = None
        self.rfe_df = None

        #The methods we will use
        self.algorithms = [
        linear_model.LogisticRegression(solver='lbfgs'),
        ensemble.RandomForestClassifier(n_estimators=100),
        ensemble.AdaBoostClassifier(),
        ensemble.GradientBoostingClassifier(),
        linear_model.SGDClassifier(loss='modified_huber'),
        svm.SVC(probability=True, gamma='scale'),
        neural_network.MLPClassifier(),
        neighbors.KNeighborsClassifier(),
        discriminant_analysis.LinearDiscriminantAnalysis(),
        discriminant_analysis.QuadraticDiscriminantAnalysis(),
        ensemble.BaggingClassifier(),
        xgboost.XGBClassifier()
        ]
    
    # Report and data summary you want 
    def summary(self):
        print("Your data looks like this (showing the first few lines of the left-most and right-most columns)...")
        print("#"*70)
        print(self.df.describe())
        print("#"*70)

    def compete(self, verbose=False):
        log_cols=["Algorithm", "AUC_Percent", "Accuracy_Percent", "Balanced_Accuracy_Percent", "Log_Loss", "Sensitivity", "Specificity", "PPV", "NPV", "Runtime_Seconds"]
        log_table = pd.DataFrame(columns=log_cols)

        for algo in self.algorithms:
            
            start_time = time.time()
            
            algo.fit(self.X_train, self.y_train)
            name = algo.__class__.__name__

            print("")
            print("#"*70)
            print("")
            print(name)

            test_predictions = algo.predict_proba(self.X_test)
            self.test_predictions = test_predictions
            test_predictions = test_predictions[:, 1]
            rocauc = metrics.roc_auc_score(self.y_test, test_predictions)
            print("AUC: {:.4%}".format(rocauc))

            test_predictions = algo.predict(self.X_test)
            acc = metrics.accuracy_score(self.y_test, test_predictions)
            print("Accuracy: {:.4%}".format(acc))

            test_predictions = algo.predict(self.X_test)
            balacc = metrics.balanced_accuracy_score(self.y_test, test_predictions)
            print("Balanced Accuracy: {:.4%}".format(balacc))
            
            CM = metrics.confusion_matrix(self.y_test, test_predictions)
            TN = CM[0][0]
            FN = CM[1][0]
            TP = CM[1][1]
            FP = CM[0][1]
            sensitivity = TP/(TP+FN)
            specificity = TN/(TN+FP)
            PPV = TP/(TP+FP)
            NPV = TN/(TN+FN)
            
            test_predictions = self.test_predictions
            #test_predictions = algo.predict_proba(self.X_test)
            #self.test_predictions = test_predictions
            ll = metrics.log_loss(self.y_test, test_predictions)
            print("Log Loss: {:.4}".format(ll))
            
            end_time = time.time()
            elapsed_time = (end_time - start_time)
            print("Runtime in seconds: {:.4}".format(elapsed_time))

            log_entry = pd.DataFrame([[name, rocauc*100, acc*100, balacc*100, ll, sensitivity, specificity, PPV, NPV, elapsed_time]], columns=log_cols)
            log_table = log_table.append(log_entry)

        print("#"*70)
        print("")

        self.log_table = log_table

        return log_table

    def results(self, metric_max):
        self.metric_max = metric_max 

        if(metric_max == "AUC"):
            best_performing_summary = self.log_table[self.log_table.AUC_Percent == self.log_table.AUC_Percent.max()]
            best_algo = best_performing_summary.at[0,'Algorithm']
        
        if(metric_max == "Balanced_Accuracy"):
            best_performing_summary = self.log_table[self.log_table.Balanced_Accuracy_Percent == self.log_table.Balanced_Accuracy_Percent.max()]
            best_algo = best_performing_summary.at[0,'Algorithm']
        
        if(metric_max == "Sensitivity"):
            best_performing_summary = self.log_table[self.log_table.Sensitivity == self.log_table.Sensitivity.max()]
            best_algo = best_performing_summary.at[0,'Algorithm']           
        
        if(metric_max == "Specificity"):
            best_performing_summary = self.log_table[self.log_table.Specificity == self.log_table.Specificity.max()]
            best_algo = best_performing_summary.at[0,'Algorithm']
        
        # If, for some reason, algorithms report the exact same score, only choose the first one so things don't crash
        if isinstance(best_algo, list):
            best_algo = best_algo[0]
        
        self.best_algo = best_algo

        return best_algo


    def AUC(self, save = False):
        plot_out = self.run_prefix + '.trainedModel_withheldSample_ROC.png'

        # Issue #24: RandomForestClassifier is finicky - can't recalculate moving forward like the other 
        if(self.best_algo == 'RandomForestClassifier'):
            test_predictions = self.test_predictions
            self.test_predictions = test_predictions
            test_predictions = test_predictions[:, 1]
        else:
            test_predictions = self.algo.predict_proba(self.X_test)
            self.test_predictions = test_predictions
            test_predictions = test_predictions[:, 1]

        fpr, tpr, thresholds = metrics.roc_curve(self.y_test, test_predictions)
        #roc_auc = metrics.auc(fpr, tpr)
        roc_auc = metrics.roc_auc_score(self.y_test, test_predictions)

        plt.figure()
        plt.plot(fpr, tpr, color='purple', label='ROC curve (area = %0.2f)' % roc_auc)
        plt.plot([0, 1], [0, 1], color='cyan', linestyle='--', label='Chance (area = %0.2f)' % 0.5)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False positive rate')
        plt.ylabel('True positive rate')
        plt.title('Receiver operating characteristic (ROC) - ' + self.best_algo)
        plt.legend(loc="lower right")
        if (save):
            plt.savefig(plot_out, dpi = 600)

        #print()
        print(f"We are also exporting a ROC curve for you here {plot_out} this is a graphical representation of AUC in the withheld test data for the best performing algorithm.")
    
    def export_prob_hist(self):
        # Exporting withheld test data
        #test_predicteds_probs = self.algo.predict_proba(self.X_test)
        test_predicteds_probs = self.test_predictions
        test_case_probs = test_predicteds_probs[:, 1]
        test_predicted_cases = self.algo.predict(self.X_test)

        test_case_probs_df = pd.DataFrame(test_case_probs)
        test_predicted_cases_df = pd.DataFrame(test_predicted_cases)
        y_test_df = pd.DataFrame(self.y_test)
        IDs_test_df = pd.DataFrame(self.IDs_test)

        test_out = pd.concat([IDs_test_df.reset_index(), y_test_df.reset_index(drop=True), test_case_probs_df.reset_index(drop=True), test_predicted_cases_df.reset_index(drop=True)], axis = 1, ignore_index=True)
        test_out.columns=['INDEX','ID',"CASE_REPORTED","CASE_PROBABILITY","CASE_PREDICTED"]
        test_out = test_out.drop(columns=['INDEX'])

        test_outfile = self.run_prefix + '.trainedModel_withheldSample_Predictions.csv'
        test_out.to_csv(test_outfile, index=False)

        print("")
        print(f"Preview of the exported predictions for the withheld test data that has been exported as {test_outfile} these are pretty straight forward.")
        print("They generally include the sample ID, the previously reported case status (1 = case), the case probability from the best performing algorithm and the predicted label from that algorithm")
        print("")
        print("#"*70)
        print(test_out.head())
        print("#"*70)


        # Exporting training data, which is by nature overfit
        train_predicteds_probs = self.algo.predict_proba(self.X_train)
        train_case_probs = train_predicteds_probs[:, 1]
        train_predicted_cases = self.algo.predict(self.X_train)

        train_case_probs_df = pd.DataFrame(train_case_probs)
        train_predicted_cases_df = pd.DataFrame(train_predicted_cases)
        y_train_df = pd.DataFrame(self.y_train)
        IDs_train_df = pd.DataFrame(self.IDs_train)

        train_out = pd.concat([IDs_train_df.reset_index(), y_train_df.reset_index(drop=True), train_case_probs_df.reset_index(drop=True), train_predicted_cases_df.reset_index(drop=True)], axis = 1, ignore_index=True)
        train_out.columns=['INDEX','ID',"CASE_REPORTED","CASE_PROBABILITY","CASE_PREDICTED"]
        train_out = train_out.drop(columns=['INDEX'])

        train_outfile = self.run_prefix + '.trainedModel_trainingSample_Predictions.csv'
        train_out.to_csv(train_outfile, index=False)

        print("")
        print(f"Preview of the exported predictions for the training samples which is naturally overfit and exported as {train_outfile} in the similar format as in the withheld test dataset that was just exported.")
        print("#"*70)
        print(train_out.head())
        print("#"*70)

        # Export historgrams of probabilities
        genoML_colors = ["cyan","purple"]

        g = sns.FacetGrid(train_out, hue="CASE_REPORTED", palette=genoML_colors, legend_out=True,)
        g = (g.map(sns.distplot, "CASE_PROBABILITY", hist=True, rug=False))
        g.add_legend()

        plot_out = self.run_prefix + '.trainedModel_withheldSample_probabilities.png'
        g.savefig(plot_out, dpi=600)

        print("")
        print("We are also exporting probability density plots to the file", plot_out, "this is a plot of the probability distributions of being a case, stratified by case and control status in the withheld test samples.")

    def export_model(self):
        best_algo = self.best_algo

        if best_algo == 'LogisticRegression':
            algo = getattr(sklearn.linear_model, best_algo)()

        elif best_algo == 'SGDClassifier':
            algo = getattr(sklearn.linear_model, best_algo)(loss='modified_huber')

        elif (best_algo == 'RandomForestClassifier') or (best_algo == 'AdaBoostClassifier') or (best_algo == 'GradientBoostingClassifier') or  (best_algo == 'BaggingClassifier'):
            algo = getattr(sklearn.ensemble, best_algo)()

        elif best_algo == 'SVC':
            algo = getattr(sklearn.svm, best_algo)(probability=True)

        elif best_algo == 'ComplementNB':
            algo = getattr(sklearn.naive_bayes, best_algo)()

        elif best_algo == 'MLPClassifier':
            algo = getattr(sklearn.neural_network, best_algo)()

        elif best_algo == 'XGBClassifier':
            algo = getattr(xgboost, best_algo)()

        elif best_algo == 'KNeighborsClassifier':
            algo = getattr(sklearn.neighbors, best_algo)()

        elif (best_algo == 'LinearDiscriminantAnalysis') or (best_algo == 'QuadraticDiscriminantAnalysis'):
            algo = getattr(sklearn.discriminant_analysis, best_algo)()

        algo.fit(self.X_train, self.y_train)
        name = algo.__class__.__name__

        print("...remember, there are occasionally slight fluctuations in model performance on the same withheld samples...")
        print("#"*70)
        print(name)

        #test_predictions = algo.predict_proba(self.X_test)
        test_predictions = self.test_predictions
        test_predictions = test_predictions[:, 1]
        rocauc = metrics.roc_auc_score(self.y_test, test_predictions)
        print("AUC: {:.4%}".format(rocauc))

        test_predictions = algo.predict(self.X_test)
        acc = metrics.accuracy_score(self.y_test, test_predictions)
        print("Accuracy: {:.4%}".format(acc))

        test_predictions = algo.predict(self.X_test)
        balacc = metrics.balanced_accuracy_score(self.y_test, test_predictions)
        print("Balanced Accuracy: {:.4%}".format(balacc))

        #test_predictions = algo.predict_proba(self.X_test)
        test_predictions = self.test_predictions
        ll = metrics.log_loss(self.y_test, test_predictions)
        print("Log Loss: {:.4}".format(ll))

        ### Save it using joblib
        algo_out = self.run_prefix + '.trainedModel.joblib'
        joblib.dump(algo, algo_out)

        print("#"*70)
        print(f"... this model has been saved as {algo_out} for later use and can be found in your working directory.")

        self.algo = algo

        return algo

    def save_results(self, run_prefix, algorithmResults = False, bestAlgorithm = False, featureRankings = False):
        if(algorithmResults):
            log_table = self.log_table
            log_outfile = self.run_prefix + '.training_withheldSamples_performanceMetrics.csv'

            print(f"This table below is also logged as {log_outfile} and is in your current working directory...")
            print("#"*70)
            print(log_table)
            print("#"*70)

            log_table.to_csv(log_outfile, index=False)

        if(bestAlgorithm):
            best_algo = self.best_algo
            print(f"Based on your withheld samples, the algorithm with the best {self.metric_max} is the {best_algo}... let's save that model for you.")
            best_algo_name_out = self.run_prefix + ".best_algorithm.txt"
            file = open(best_algo_name_out,'w')
            file.write(self.best_algo)
            file.close() 



# Copyright 2020 The GenoML Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import sys
import joblib
import pandas as pd

from genoml.discrete import supervised

def main(prefix, test_prefix, refModel_prefix):
    print("")
    print("Here is some basic info on the command you are about to run.")
    print("Python version info...")
    print(sys.version)

    # Print out the chosen CLI arguments 
    print("CLI argument info...")
    print(f"You are importing this test dataset: {test_prefix}.")
    print(f"You are applying the model saved here: {refModel_prefix}.")
    print(f"The results of this test application of your model will be saved in files with the given prefix: {prefix}.")
    print("As a note, in all exported probabilities and other graphics, case status is treated as a 0 or 1, with 1 representing a positive case.")

    print("")

    # Specify prefix and dataframe variables to be passed into class
    run_prefix = prefix
    infile_h5 = test_prefix + ".dataForML.h5"
    df = pd.read_hdf(infile_h5, key = "dataForML")

    infile_model = refModel_prefix + ".joblib"
    loaded_model = joblib.load(infile_model)

    # Pass the arguments to the class 
    test = supervised.test(df, loaded_model, run_prefix)

    # Prep and show the dataframe
    test.prep_df()

    # Export the ROC
    test.export_ROC()

    # Export the tested data 
    test.export_tested_data()

    # Export the histograms
    test.export_histograms()

    # Export the additional summary stats
    test.additional_sumstats()

    # Thank the user
    print("")
    print("Let's shut everything down, thanks for testing your model with GenoML!")
    print("")

    
